from __future__ import annotations

import logging
import re
import uuid
from typing import TYPE_CHECKING, Callable

from flow.record.fieldtypes import posix_path

from dissect.target.exceptions import FilesystemError
from dissect.target.filesystems.nfs import NfsFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.nfs.client.nfs import Client as NfsClient
from dissect.target.helpers.nfs.client.nfs import NfsError
from dissect.target.helpers.nfs.nfs3 import FileHandle, NfsStat
from dissect.target.helpers.record import UnixUserRecord
from dissect.target.helpers.sunrpc.client import LocalPortPolicy, auth_unix
from dissect.target.helpers.utils import parse_options_string
from dissect.target.loaders.local import LocalLoader
from dissect.target.plugin import OperatingSystem, OSPlugin, arg, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target

log = logging.getLogger(__name__)


# https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#ISA
ARCH_MAP = {
    0x00: "unknown",
    0x02: "sparc",
    0x03: "x86",
    0x08: "mips",
    0x14: "powerpc32",
    0x15: "powerpc64",
    0x16: "s390",  # and s390x
    0x28: "aarch32",  # armv7
    0x2A: "superh",
    0x32: "ia-64",
    0x3E: "x86_64",
    0xB7: "aarch64",  # armv8
    0xF3: "riscv64",
    0xF7: "bpf",
}


class UnixPlugin(OSPlugin):
    """UNIX plugin."""

    # Files to parse for user details
    PASSWD_FILES = ("/etc/passwd", "/etc/passwd-", "/etc/master.passwd")

    def __init__(self, target: Target):
        super().__init__(target)
        self._add_mounts()
        self._add_devices()
        self._hostname, self._domain = self._parse_hostname_string()
        self._hosts_dict = self._parse_hosts_string()
        self._os_release = self._parse_os_release()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        # If the ``/var`` and ``/etc`` folders exist on a filesystem it is treated as a Unix-like filesystem
        for fs in target.filesystems:
            if fs.exists("/var") and fs.exists("/etc"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(record=UnixUserRecord)
    @arg("--sessions", action="store_true", help="Parse syslog for recent user sessions")
    def users(self, sessions: bool = False) -> Iterator[UnixUserRecord]:
        """Yield unix user records from passwd files or syslog session logins.

        References:
            - https://manpages.ubuntu.com/manpages/oracular/en/man5/passwd.5.html
        """

        seen_users = set()

        # Yield users found in passwd files.
        for passwd_file in self.PASSWD_FILES:
            if (path := self.target.fs.path(passwd_file)).exists():
                for line in path.open("rt", errors="surrogateescape"):
                    line = line.strip()
                    if not line or line.startswith("#"):
                        continue

                    pwent = dict(enumerate(line.split(":")))

                    current_user = (pwent.get(0), pwent.get(5), pwent.get(6))
                    if current_user in seen_users:
                        continue
                    seen_users.add(current_user)
                    yield UnixUserRecord(
                        name=pwent.get(0),
                        passwd=pwent.get(1),
                        uid=pwent.get(2) or None,
                        gid=pwent.get(3) or None,
                        gecos=pwent.get(4),
                        home=posix_path(pwent.get(5)),
                        shell=pwent.get(6),
                        source=passwd_file,
                        _target=self.target,
                    )

        # Find users not in passwd files by parsing recent
        # syslog ldap, kerberos and x-session logins.
        # Must be enabled using the --sessions flag
        if sessions and (path := self.target.fs.path("/var/log/syslog")).exists():
            sessions = []
            cur_session = -1
            needles = {
                ("setting HOME=", "home"),
                ("setting SHELL=", "shell"),
                ("setting USER=", "name"),
            }

            for line in path.open("rt"):
                # Detect the beginning of a new session activation in the syslog
                #
                # dbus-update-activation-environment starts a new session with
                # DBUS_SESSION_BUS_ADDRESS, XDG_RUNTIME_DIR and/or DISPLAY
                # Using DBUS_SESSION_BUS_ADDRESS seems to work fine for now.
                if "setting DBUS_SESSION_BUS_ADDRESS=" in line:
                    bus = line.split("DBUS_SESSION_BUS_ADDRESS=")[1].strip()
                    cur_session += 1
                    sessions.append({"bus": bus, "name": None, "home": None, "shell": None, "uid": None})

                # Search the following lines for more information on the previousley detected session.
                for n, k in needles:
                    if n in line and "dbus-update-activation-environment" in line:
                        sessions[cur_session][k] = line.split(n)[1].strip()

            for user in sessions:
                # Only return users we didn't already find in previously parsed passwd files and past sessions.
                current_user = (user["name"], user["home"], user["shell"])
                if current_user in seen_users:
                    continue

                seen_users.add(current_user)

                yield UnixUserRecord(
                    name=user["name"],
                    home=posix_path(user["home"]),
                    shell=user["shell"],
                    source="/var/log/syslog",
                    _target=self.target,
                )

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        if (root_path := self.target.fs.path("/root")).exists():
            yield (root_path, ("uid", 0))

        if (home_path := self.target.fs.path("/home")).exists():
            yield from ((entry, None) for entry in home_path.iterdir() if entry.is_dir())

    @export(property=True)
    def architecture(self) -> str | None:
        return self._get_architecture(self.os)

    @export(property=True)
    def hostname(self) -> str | None:
        return self._hostname or self._hosts_dict.get("hostname", "localhost")

    @export(property=True)
    def domain(self) -> str | None:
        if self._domain and "localhost" not in self._domain:
            return self._domain

        # fall back to /etc/hosts file
        if "localhost" not in (domain := self._hosts_dict.get("hostname", "")):
            return domain or None

        return None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.UNIX.value

    def _parse_hostname_string(
        self, paths: list[tuple[str, Callable[[Path], str] | None]] | None = None
    ) -> tuple[str | None, str | None]:
        """Returns a tuple containing respectively the hostname and domain name portion of the path(s) specified.

        Args:
            paths (list): list of tuples with paths and callables to parse the path or None

        Returns:
            Tuple with hostname and domain strings.
        """
        hostname = None
        domain = None

        paths = paths or [
            ("/etc/hostname", None),
            ("/etc/HOSTNAME", None),
            ("/proc/sys/kernel/hostname", None),
            ("/etc/sysconfig/network", self._parse_rh_legacy),
            ("/etc/hosts", self._parse_etc_hosts),  # fallback if no other hostnames are found
        ]

        for path, callable in paths:
            if not (path := self.target.fs.path(path)).exists():
                continue

            hostname = callable(path) if callable else path.open("rt").read().rstrip()

            if hostname and "." in hostname:
                hostname, domain = hostname.split(".", maxsplit=1)

            break  # break whenever a valid hostname is found

        # Can be an empty string due to splitting of hostname and domain
        return hostname or None, domain or None

    def _parse_rh_legacy(self, path: Path) -> str | None:
        hostname = None
        file_contents = path.open("rt").readlines()
        for line in file_contents:
            if not line.startswith("HOSTNAME"):
                continue
            _, _, hostname = line.rstrip().partition("=")
        return hostname

    def _parse_etc_hosts(self, path: Path) -> str | None:
        for line in path.open("rt"):
            if line.startswith(("127.0.0.1 ", "::1 ")) and "localhost" not in line:
                return line.split(" ")[1]
        return None

    def _parse_hosts_string(self, paths: list[str] | None = None) -> dict[str, str]:
        paths = paths or ["/etc/hosts"]
        hosts_string = {}

        for path in paths:
            for fs in self.target.filesystems:
                if fs.exists(path):
                    for line in fs.path(path).open("rt").readlines():
                        if not (line := line.split()):
                            continue
                        if line[0].startswith(("127.0.", "::1")):
                            hosts_string = {"ip": line[0], "hostname": line[1]}
        return hosts_string

    def _add_mounts(self) -> None:
        fstab = self.target.fs.path("/etc/fstab")

        for dev_id, volume_name, mount_point, fs_type, options in parse_fstab(fstab, self.target.log):
            # Mount nfs, but only when target has been mapped by the `LocalLoader`
            if fs_type == "nfs":
                self._add_nfs(dev_id, volume_name, mount_point)
                continue

            opts = parse_options_string(options)
            subvol = opts.get("subvol", None)
            subvolid = opts.get("subvolid", None)
            for fs in self.target.filesystems:
                fs_id = None
                fs_subvol = None
                fs_subvolid = None
                fs_last_mount = None
                fs_volume_name = None
                vol_volume_name = fs.volume.name if fs.volume and not isinstance(fs.volume, list) else None

                if fs.__type__ == "xfs":
                    fs_id = fs.xfs.uuid
                    fs_volume_name = fs.xfs.name
                elif fs.__type__ == "ext":
                    fs_id = fs.extfs.uuid
                    fs_last_mount = fs.extfs.last_mount
                    fs_volume_name = fs.extfs.volume_name
                elif fs.__type__ == "btrfs":
                    fs_id = fs.btrfs.uuid
                    fs_subvol = fs.subvolume.path
                    fs_subvolid = fs.subvolume.objectid
                elif fs.__type__ == "fat":
                    fs_id = fs.fatfs.volume_id
                    # This normalizes fs_id to comply with libblkid generated UUIDs
                    # This is needed because FAT filesystems don't have a real UUID,
                    # but instead a volume_id which is not case-sensitive
                    fs_id = fs_id[:4].upper() + "-" + fs_id[4:].upper()

                if (
                    (fs_id and (fs_id == dev_id and (subvol == fs_subvol or subvolid == fs_subvolid)))
                    or (fs_last_mount and (fs_last_mount == mount_point))
                    or (fs_volume_name and (fs_volume_name == volume_name))
                    or (vol_volume_name and (vol_volume_name == volume_name))
                ):
                    self.target.log.debug("Mounting %s (%s) at %s", fs, fs.volume, mount_point)
                    self.target.fs.mount(mount_point, fs)

    @property
    def _is_nfs_enabled(self) -> bool:
        return isinstance(self.target._loader, LocalLoader) and "enable-nfs" in self.target.path_query

    def _log_nfs_mount_disabled(self, address: str, exported_dir: str, mount_point: str) -> None:
        if isinstance(self.target._loader, LocalLoader):
            log.warning(
                "NFS mount %s:%s at %s is disabled. To enable, pass --enable-nfs to the local loader. Alternatively, add a query parameter to the target query string: local?enable-nfs",  # noqa: E501
                address,
                exported_dir,
                mount_point,
            )
        else:
            log.warning(
                "NFS mount %s:%s at %s is unavailable on a non-local target",
                address,
                exported_dir,
                mount_point,
            )

    def _add_nfs(self, address: str, exported_dir: str, mount_point: str) -> None:
        if not self._is_nfs_enabled:
            self._log_nfs_mount_disabled(address, exported_dir, mount_point)
            return

        # Try all users to see if we can access the share
        def auth_setter(nfs_client: NfsClient, filehandle: FileHandle, _: list[int]) -> None:
            users = list(self.users())
            if not users:
                self.target.log.debug(
                    "No users found, trying root for mounting NFS share %s:%s at %s", address, exported_dir, mount_point
                )
                users = [UnixUserRecord(uid=0, gid=0)]

            for user in users:
                if user.uid is None or user.gid is None:
                    continue
                auth = auth_unix("machine", user.uid, user.gid, [])
                nfs_client.rebind_auth(auth)
                try:
                    self.target.log.debug("Trying to read NFS share with uid %d and gid %d", user.uid, user.gid)
                    # Use a readdir to check if we have access.
                    # RdJ: Perhaps an ACCESS call (to be implemented) is better than READDIR
                    nfs_client.readdir(filehandle)
                except NfsError as e:
                    if e.nfsstat != NfsStat.ERR_ACCES:
                        self.target.log.warning("Reading NFS share gives %s", e.nfsstat)
                        nfs_client.close()
                        raise
                else:
                    # We have access
                    return

            self.target.log.debug("No user has access to NFS share %s:%s at %s", address, exported_dir, mount_point)
            raise FilesystemError(f"No user has access to NFS share {address}:{exported_dir} at {mount_point}")

        try:
            self.target.log.debug("Mounting NFS share %s at %s", exported_dir, mount_point)
            nfs = NfsFilesystem.connect(address, exported_dir, auth_setter, LocalPortPolicy.PRIVILEGED)
            self.target.fs.mount(mount_point, nfs)
        except Exception as e:
            self.target.log.warning("Failed to mount NFS share %s:%s at %s", address, exported_dir, mount_point)
            self.target.log.debug("", exc_info=e)

    def _add_devices(self) -> None:
        """Add some virtual block devices to the target.

        Currently only adds LVM devices.
        """
        vfs = self.target.fs.append_layer()

        for volume in self.target.volumes:
            if volume.vs and volume.vs.__type__ == "lvm":
                vfs.map_file_fh(f"/dev/{volume.raw.vg.name}/{volume.raw.name}", volume)

    def _parse_os_release(self, glob: str | None = None) -> dict[str, str]:
        """Parse files containing Unix version information.

        Not all these files are equal. Generally speaking these files are
        either key=value files or contain just one line.

        Examples of key=value pair structured release files are:
        - /etc/os-release
        - /usr/lib/os-release
        - /etc/lsb-release

        Examples of sparse release files:
        - /etc/fedora-release
        - /etc/centos-release
        - /etc/redhat-release
        - /etc/SuSE-release

        Examples of bsd version files:
        - /bin/freebsd-version
        """
        glob = glob or "/etc/*-release"

        os_release = {}

        for path in self.target.fs.glob(glob):
            if self.target.fs.path(path).is_file():
                with self.target.fs.path(path).open("rt") as release_file:
                    for line in release_file:
                        if line.startswith("#"):
                            continue

                        if "=" not in line:
                            os_release["DISTRIB_DESCRIPTION"] = line.strip()

                        else:
                            try:
                                name, value = line.split("=", maxsplit=1)
                                os_release[name] = value.replace('"', "").replace("\n", "")
                            except ValueError:
                                continue
        return os_release

    def _get_architecture(self, os: str = "unix", path: Path | str = "/bin/ls") -> str | None:
        """Determine architecture by reading an ELF header of a binary on the target.

        References:
            - https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#ISA
        """

        if not isinstance(path, TargetPath):
            for fs in [self.target.fs, *self.target.filesystems]:
                if (path := fs.path(path)).exists():
                    break

        if not path.exists():
            return None

        fh = path.open("rb")
        fh.seek(4)  # ELF - e_ident[EI_CLASS]
        bits = fh.read(1)[0]

        fh.seek(18)  # ELF - e_machine
        e_machine = int.from_bytes(fh.read(2), "little")
        arch = ARCH_MAP.get(e_machine, "unknown")

        return f"{arch}_32-{os}" if bits == 1 and arch[-2:] != "32" else f"{arch}-{os}"


def parse_fstab(
    fstab: TargetPath,
    log: logging.Logger = log,
) -> Iterator[tuple[uuid.UUID | str, str, str, str, str]]:
    """Parse fstab file and return a generator that streams the details of entries,
    with unsupported FS types and block devices filtered away.
    """

    SKIP_FS_TYPES = (
        "swap",
        "tmpfs",
        "devpts",
        "sysfs",
        "procfs",
        "overlayfs",
    )

    if not fstab.exists():
        return

    for entry in fstab.open("rt"):
        entry = entry.strip()
        if entry.startswith("#"):
            continue

        entry_parts = re.split(r"\s+", entry)

        if len(entry_parts) != 6:
            continue

        dev, mount_point, fs_type, options, _, _ = entry_parts

        if fs_type in SKIP_FS_TYPES:
            log.warning("Skipped FS type: %s, %s, %s", fs_type, dev, mount_point)
            continue

        dev_id = None
        volume_name = None
        if dev.startswith(("/dev/mapper", "/dev/gpt")):
            volume_name = dev.rsplit("/")[-1]
        elif dev.startswith("/dev/disk/by-uuid"):
            dev_id = dev.rsplit("/")[-1]
        elif dev.startswith("/dev/") and dev.count("/") == 3:
            # When composing a vg-lv name, LVM2 replaces hyphens with double hyphens in the vg and lv names
            # Emulate that here when combining the vg and lv names
            volume_name = "-".join(part.replace("-", "--") for part in dev.rsplit("/")[-2:])
        elif dev.startswith("UUID="):
            dev_id = dev.split("=")[1]
        elif dev.startswith("LABEL="):
            volume_name = dev.split("=")[1]
        elif fs_type == "nfs":
            # Put the nfs server address in dev_id and the root path in volume_name
            dev_id, sep, volume_name = dev.partition(":")
            if sep != ":":
                log.warning("Invalid NFS mount: %s", dev)
                continue
        else:
            log.warning("Unsupported mount device: %s %s", dev, mount_point)
            continue

        if mount_point == "/":
            continue

        if dev_id:
            try:
                dev_id = uuid.UUID(dev_id)
            except ValueError:
                pass

        yield dev_id, volume_name, mount_point, fs_type, options



from __future__ import annotations

import hashlib
import plistlib
import struct
from io import BytesIO
from typing import TYPE_CHECKING, Any

from dissect.sql import sqlite3
from dissect.util.plist import NSKeyedArchiver

from dissect.target.exceptions import LoaderError
from dissect.target.filesystems.itunes import ITunesFilesystem
from dissect.target.helpers import fsutil, keychain
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.target import Target

try:
    import _pystandalone

    HAS_PYSTANDALONE = True
except ImportError:
    HAS_PYSTANDALONE = False

try:
    from Crypto.Cipher import AES

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False


DOMAIN_TRANSLATION = {
    "AppDomain": "/private/var/mobile/Containers/Data/Application",
    "AppDomainGroup": "/private/var/mobile/Containers/Shared/AppGroup",
    "AppDomainPlugin": "/private/var/mobile/Containers/Data/PluginKitPlugin",
    "SysContainerDomain": "/private/var/containers/Data/System",
    "SysSharedContainerDomain": "/private/var/containers/Shared/SystemGroup",
    "KeychainDomain": "/private/var/Keychains",
    "CameraRollDomain": "/private/var/mobile",
    "MobileDeviceDomain": "/private/var/MobileDevice",
    "WirelessDomain": "/private/var/wireless",
    "InstallDomain": "/private/var/installd",
    "KeyboardDomain": "/private/var/mobile",
    "HomeDomain": "/private/var/mobile",
    "SystemPreferencesDomain": "/private/var/preferences",
    "DatabaseDomain": "/private/var/db",
    "TonesDomain": "/private/var/mobile",
    "RootDomain": "/private/var/root",
    "BooksDomain": "/private/var/mobile/Media/Books",
    "ManagedPreferencesDomain": "/private/var/Managed Preferences",
    "HomeKitDomain": "/private/var/mobile",
    "MediaDomain": "/private/var/mobile",
    "HealthDomain": "/private/var/mobile/Library",
}


class ITunesLoader(Loader):
    """Load iTunes backup files.

    References:
        - https://support.apple.com/en-us/HT204215
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        self.backup = ITunesBackup(self.path)

        if self.backup.encrypted:
            for key in keychain.get_keys_for_provider("itunes") + keychain.get_keys_without_provider():
                if key.key_type == keychain.KeyType.PASSPHRASE:
                    if key.identifier == self.backup.identifier:
                        self.backup.open(key.value)
                        break
                    if not key.identifier:
                        try:
                            self.backup.open(key.value)
                            break
                        except ValueError:
                            continue
            else:
                raise LoaderError(f"No password for encrypted iTunes backup: {self.path}")
        else:
            self.backup.open()

    @staticmethod
    def detect(path: Path) -> bool:
        return path.is_dir() and path.joinpath("Manifest.plist").exists()

    def map(self, target: Target) -> None:
        target.filesystems.add(ITunesFilesystem(self.backup))


class ITunesBackup:
    """Parse a directory as an iTunes backup directory."""

    def __init__(self, root: Path):
        self.root = root
        self.manifest = plistlib.load(self.root.joinpath("Manifest.plist").open("rb"))
        self.info = plistlib.load(self.root.joinpath("Info.plist").open("rb"))
        self.status = plistlib.load(self.root.joinpath("Status.plist").open("rb"))

        self.encrypted = self.manifest["IsEncrypted"]

        self.kek = None
        self.key_bag = None

        if self.encrypted:
            self.key_bag = KeyBag(self.manifest["BackupKeyBag"])

        self.manifest_db = None

    @property
    def identifier(self) -> str:
        return self.info["Unique Identifier"]

    def open(self, password: str | None = None, kek: bytes | None = None) -> None:
        """Open the backup.

        Opens the Manifest.db file. Requires a password if the backup is encrypted.

        Args:
            password: Optional backup password if the backup is encrypted.
            kek: Optional kek if the password is unknown, but the derived key is known.
        """
        if self.encrypted:
            if not password and not kek:
                raise ValueError("Either password or kek is required for encrypted backups")

            self.kek = kek
            if not self.kek:
                self.kek = self.derive_key(password)

            self.key_bag.unlock_with_passcode_key(self.kek)

        self.manifest_db = self._open_manifest_db()

    def _open_manifest_db(self) -> sqlite3.SQLite3:
        path = self.root.joinpath("Manifest.db")
        if not self.encrypted or self.manifest["Lockdown"]["ProductVersion"] < "10.2":
            fh = path.open("rb")
        else:
            key = self.key_bag.unwrap(self.manifest["ManifestKey"])
            fh = BytesIO(aes_decrypt(path.read_bytes(), key))

        return sqlite3.SQLite3(fh)

    def derive_key(self, password: str) -> bytes:
        """Derive the key bag encryption key from a given password."""
        password = password.encode()
        if self.manifest["Lockdown"]["ProductVersion"] < "10.2":
            first_round = password
        else:
            first_round = hashlib.pbkdf2_hmac(
                "sha256", password, self.key_bag.attr["DPSL"], self.key_bag.attr["DPIC"], 32
            )

        return hashlib.pbkdf2_hmac("sha1", first_round, self.key_bag.attr["SALT"], self.key_bag.attr["ITER"], 32)

    def files(self) -> Iterator[FileInfo]:
        """Iterate all the files in this backup."""

        if table := self.manifest_db.table("Files"):
            for row in table.rows():
                yield FileInfo(self, row.fileID, row.domain, row.relativePath, row.flags, row.file)


class FileInfo:
    """Utility class that represents a file in a iTunes backup."""

    def __init__(
        self, backup: ITunesBackup, file_id: str, domain: str, relative_path: str, flags: int, metadata: bytes
    ):
        self.backup = backup
        self.file_id = file_id
        self.domain = domain
        self.relative_path = relative_path
        self.flags = flags
        self.metadata = NSKeyedArchiver(BytesIO(metadata))["root"]

        self.translated_path = translate_file_path(self.domain, self.relative_path)

    def __repr__(self) -> str:
        return f"<FileInfo {self.translated_path}>"

    @property
    def mode(self) -> int:
        return self.metadata["Mode"]

    @property
    def size(self) -> int:
        return self.metadata["Size"]

    @property
    def encryption_key(self) -> str | None:
        return self.metadata.get("EncryptionKey")

    def get(self) -> Path:
        """Return a Path object to the underlying file."""
        return self.backup.root / self.file_id[:2] / self.file_id

    def create_cipher(self) -> Any:
        """Return a new AES cipher for this file."""
        if not self.backup.encrypted:
            raise TypeError("File is not encrypted")

        if not self.encryption_key:
            raise ValueError("File has no encryption key")

        key = self.backup.key_bag.unwrap(self.encryption_key)
        return _create_cipher(key)


class KeyBag:
    """Parse and implements a simple key bag."""

    def __init__(self, buf: bytes):
        self.attr, self.keys = parse_key_bag(buf)

    def unlock_with_passcode_key(self, key: bytes) -> None:
        """Attempt to unlock the passcode protected keys in this key bag with the given decryption key."""
        for class_key in self.keys.values():
            if class_key.wrap_type != ClassKey.WRAP_PASSCODE:
                continue

            class_key.unwrap(key)

    def unwrap(self, key: bytes) -> bytes:
        """Unwrap a given key.

        Wrapped keys are prefixed with a 32bit protection class.
        """
        protection_class, wrapped_key = struct.unpack("<I", key[:4])[0], key[4:]
        return aes_unwrap_key(self.keys[protection_class].key, wrapped_key)


class ClassKey:
    """Represent a class key that is stored in a key bag."""

    WRAP_PASSCODE = 2

    def __init__(
        self,
        uuid: bytes,
        protection_class: int,
        wrap_type: int,
        key_type: int,
        wrapped_key: bytes,
        public_key: bytes | None = None,
    ):
        self.uuid = uuid
        self.protection_class = protection_class
        self.wrap_type = wrap_type
        self.key_type = key_type
        self.wrapped_key = wrapped_key
        self.public_key = public_key

        self.key = None

    @classmethod
    def from_bag_dict(cls, data: dict[str, bytes | int]) -> Self:
        return cls(
            data.get("UUID"),
            data.get("CLAS"),
            data.get("WRAP"),
            data.get("KTYP"),
            data.get("WPKY"),
            data.get("PBKY"),
        )

    @property
    def unwrapped(self) -> bool:
        """Return whether this key is already unwrapped."""
        return self.key is not None

    def unwrap(self, kek: bytes) -> None:
        """Attempt to unwrap this key."""
        self.key = aes_unwrap_key(kek, self.wrapped_key)


def translate_file_path(domain: str, relative_path: str) -> str:
    """Translate a domain and relative path (as stored in iTunes backups) to an absolute path on an iOS device."""
    try:
        domain, _, package_name = domain.partition("-")
    except ValueError:
        package_name = ""

    domain_path = fsutil.join(DOMAIN_TRANSLATION.get(domain, domain), package_name)
    return fsutil.join(domain_path, relative_path).rstrip("/")


def parse_key_bag(buf: bytes) -> tuple[dict[str, bytes, int], dict[str, ClassKey]]:
    """Parse the BackupKeyBag buffer. Simple TLV format."""
    attr = {}
    class_keys = {}
    current_class_key = {}

    fh = BytesIO(buf)
    while True:
        header = fh.read(8)
        if len(header) != 8:
            break

        key, length = struct.unpack(">4sI", header)

        key = key.decode()
        value = fh.read(length)
        if length == 4:
            value = struct.unpack(">I", value)[0]

        # The order here is important
        # There are some fields which qualify as a "header" of the key bag, which we name attributes
        # Among these "header" fields is a UUID field that acts as the UUID for the key bag
        # After this header comes a list of class keys. Each class key starts with a UUID
        # We only want to start parsing class keys from the second UUID we encounter, because
        # the first UUID we encounter will be from the "header"
        if key == "UUID" and "UUID" in attr:
            if current_class_key:
                ckey = ClassKey.from_bag_dict(current_class_key)
                class_keys[ckey.protection_class] = ckey

            current_class_key = {
                "UUID": value,
            }
        elif current_class_key and key in ("CLAS", "WRAP", "KTYP", "WPKY", "PKBY"):
            current_class_key[key] = value
        else:
            attr[key] = value

    if current_class_key:
        ckey = ClassKey.from_bag_dict(current_class_key)
        class_keys[ckey.protection_class] = ckey

    return attr, class_keys


def aes_decrypt(data: bytes, key: bytes, iv: bytes = b"\x00" * 16) -> bytes:
    """Helper function to easily decrypt some data with a default IV."""
    return _create_cipher(key, iv).decrypt(data)


def aes_unwrap_key(kek: bytes, wrapped: bytes, iv: int = 0xA6A6A6A6A6A6A6A6) -> bytes:
    """AES key unwrapping algorithm.

    Derived from https://github.com/kurtbrose/aes_keywrap/blob/master/aes_keywrap.py
    """
    QUAD = struct.Struct(">Q")

    n = len(wrapped) // 8 - 1

    # NOTE: R[0] is never accessed, left in for consistency with RFC indices
    R = [None] + [wrapped[i * 8 : i * 8 + 8] for i in range(1, n + 1)]
    A = QUAD.unpack(wrapped[:8])[0]

    decrypt = _create_cipher(kek, mode="ecb").decrypt

    for j in range(5, -1, -1):  # counting down
        for i in range(n, 0, -1):  # (n, n-1, ..., 1)
            ciphertext = QUAD.pack(A ^ (n * j + i)) + R[i]
            B = decrypt(ciphertext)
            A = QUAD.unpack(B[:8])[0]
            R[i] = B[8:]

    key, key_iv = b"".join(R[1:]), A

    if key_iv != iv:
        raise ValueError(f"Unwrapping failed: 0x{key_iv:x} (expected 0x{iv:x})")

    return key


def _create_cipher(key: bytes, iv: bytes = b"\x00" * 16, mode: str = "cbc") -> Any:
    """Create a cipher object.

    Dynamic based on the available crypto module.
    """

    if HAS_PYSTANDALONE:
        key_size = len(key)
        if key_size not in (32, 24, 16):
            raise ValueError(f"Invalid key size: {key_size}")

        return _pystandalone.cipher(f"aes-{key_size * 8}-{mode}", key, iv)
    if HAS_CRYPTO:
        mode_map = {
            "cbc": (AES.MODE_CBC, True),
            "ecb": (AES.MODE_ECB, False),
        }
        mode_id, has_iv = mode_map[mode]
        kwargs = {"iv": iv} if has_iv else {}
        return AES.new(key, mode_id, **kwargs)
    raise RuntimeError("No crypto module available")



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, alias, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

VmwareDragAndDropRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "virtualization/vmware/clipboard",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

VMWARE_DND_PATHS = [
    # Windows
    "AppData/Local/Temp/VmwareDND",
    # Linux
    ".cache/vmware/drag_and_drop",
]


class VmwareWorkstationPlugin(Plugin):
    """VMware Workstation plugin."""

    __namespace__ = "vmware"

    def __init__(self, target: Target):
        super().__init__(target)
        self.dnd_dirs = list(self.find_dnd_dirs())

    def check_compatible(self) -> None:
        if not self.dnd_dirs:
            raise UnsupportedPluginError("No VMware Workstation DnD artifact(s) found")

    def find_dnd_dirs(self) -> Iterator[tuple[UserDetails, TargetPath]]:
        for user_details in self.target.user_details.all_with_home():
            for dnd_path in VMWARE_DND_PATHS:
                if (dnd_dir := user_details.home_path.joinpath(dnd_path)).exists():
                    yield user_details, dnd_dir

    @alias("draganddrop")
    @export(record=VmwareDragAndDropRecord)
    def clipboard(self) -> Iterator[VmwareDragAndDropRecord]:
        """Yield cached VMware Workstation drag-and-drop file artifacts."""

        for user_details, dnd_dir in self.dnd_dirs:
            for file in dnd_dir.rglob("*/*"):
                if file.is_dir():
                    continue

                yield VmwareDragAndDropRecord(
                    ts=file.lstat().st_mtime,
                    path=file,
                    _user=user_details.user,
                    _target=self.target,
                )



from __future__ import annotations

import contextlib
import json
import re
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest

from dissect.target.plugin import FunctionDescriptor, Plugin, PluginRegistry, arg, export
from dissect.target.tools.query import main as target_query
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_list(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-query", "--list"])

        with pytest.raises(SystemExit):
            target_query()

        out, _ = capsys.readouterr()

        assert out.startswith("Available plugins:")
        assert "Failed to load:\n    None\n\nAvailable loaders:\n" in out


@pytest.mark.parametrize(
    ("target_fixture"),
    [
        "target_win_users",
        "target_unix_users",
        "target_linux_users",
        "target_macos_users",
    ],
)
def test_list_target(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
    request: pytest.FixtureRequest,
    target_fixture: str,
) -> None:
    """Tests whether ``--list`` and ``--list *`` on a target returns the same results."""

    args = ["target-query", "mock/path", "--list"]

    target: Target = request.getfixturevalue(target_fixture)

    def _run_query(args: list[str], target: Target) -> tuple[bytes, bytes]:
        with monkeypatch.context() as m:
            m.setattr("sys.argv", args)

            # Patch the target that gets opened.
            with patch("dissect.target.target.Target.open_all", return_value=[target]), contextlib.suppress(SystemExit):
                target_query()

            return capsys.readouterr()

    stdout_1, stderr_1 = _run_query(args, target)
    stdout_2, stderr_2 = _run_query([*args, "*"], target)

    assert stdout_1 == stdout_2
    assert stderr_1 == stderr_2


@pytest.mark.parametrize(
    ("given_funcs", "expected_invalid_funcs"),
    [
        (
            ["foo"],
            ["foo"],
        ),
        (
            ["bar", "version"],
            ["bar"],
        ),
        (
            ["version", "foo", "wireguard.config", "bar", "apps.webserver.iis*"],
            ["bar", "foo"],
        ),
        (
            ["apps.browser.*.downloads", "bar", "version", "foo"],
            ["bar", "foo"],
        ),
        (
            ["apps.webserver.iis.doesnt.exist", "apps.webserver.apache.access"],
            ["apps.webserver.iis.doesnt.exist"],
        ),
    ],
)
def test_invalid_functions(
    given_funcs: list[str],
    expected_invalid_funcs: list[str],
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", ",".join(given_funcs), str(absolute_path("_data/loaders/tar/test-archive.tar.gz"))],
        )

        with pytest.raises(SystemExit):
            target_query()
        _, err = capsys.readouterr()

        assert "target-query: error: argument -f/--function contains invalid plugin(s):" in err

        # Workaround for https://github.com/fox-it/dissect.target/issues/266
        RE_ERR = re.compile(r"contains invalid plugin\(s\)\: (?P<funcs>.*?)\n$")
        match = RE_ERR.search(err).groupdict()

        invalid_funcs = [i.strip() for i in match["funcs"].split(",")]
        invalid_funcs.sort()

        assert invalid_funcs == expected_invalid_funcs


@pytest.mark.parametrize(
    ("given_funcs", "expected_invalid_funcs"),
    [
        (
            ["foo"],
            ["foo"],
        ),
        (
            ["bar", "version"],
            ["bar"],
        ),
        (
            ["version", "foo", "wireguard.config", "bar", "apps.webserver.iis*"],
            ["bar", "foo"],
        ),
        (
            ["apps.browser.*.downloads", "bar", "version", "foo"],
            ["bar", "foo"],
        ),
        (
            ["apps.webserver.iis.doesnt.exist", "apps.webserver.apache.access"],
            ["apps.webserver.iis.doesnt.exist"],
        ),
    ],
)
def test_invalid_excluded_functions(
    given_funcs: list[str],
    expected_invalid_funcs: list[str],
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-f",
                "hostname",
                "-xf",
                ",".join(given_funcs),
                str(absolute_path("_data/loaders/tar/test-archive.tar.gz")),
            ],
        )

        with pytest.raises(SystemExit):
            target_query()
        _, err = capsys.readouterr()

        assert "target-query: error: argument -xf/--excluded-functions contains invalid plugin(s):" in err

        # Workaround for https://github.com/fox-it/dissect.target/issues/266
        RE_ERR = re.compile(r"contains invalid plugin\(s\)\: (?P<funcs>.*?)\n$")
        match = RE_ERR.search(err).groupdict()

        invalid_funcs = [i.strip() for i in match["funcs"].split(",")]
        invalid_funcs.sort()

        assert invalid_funcs == expected_invalid_funcs


def test_unsupported_plugin_log(caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", "regf", str(absolute_path("_data/loaders/tar/test-archive.tar.gz"))],
        )

        target_query()

        assert "Unsupported plugin for regf: Registry plugin not loaded" in caplog.text


def mock_find_functions(patterns: str, *args, **kwargs) -> tuple[list[FunctionDescriptor], set[str]]:
    plugins = [
        FunctionDescriptor(
            name=pattern,
            namespace=None,
            path=pattern,
            exported=True,
            internal=False,
            findable=True,
            alias=False,
            output="record",
            method_name=pattern,
            module=pattern,
            qualname=pattern.capitalize(),
        )
        for pattern in patterns.split(",")
    ]

    return (plugins, set())


def mock_execute_function(
    target: Target,
    func: FunctionDescriptor,
    arguments: list[str] | None = None,
) -> tuple[str, Any, list[str]]:
    return (func.output, func.name)


def test_filtered_functions(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-f",
                "foo,bar,bla,foo",
                "-xf",
                "bla",
                str(absolute_path("_data/loaders/tar/test-archive.tar.gz")),
            ],
        )

        with (
            patch(
                "dissect.target.tools.utils.find_functions",
                autospec=True,
                side_effect=mock_find_functions,
            ),
            patch(
                "dissect.target.tools.query.execute_function_on_target",
                autospec=True,
                side_effect=mock_execute_function,
            ) as mock_execute,
            patch(
                "dissect.target.tools.query.record_output",
                autospec=True,
            ),
        ):
            target_query()

            assert len(mock_execute.mock_calls) == 2

            executed_func_names = set()
            for call in mock_execute.mock_calls:
                executed_func_names.add(call.args[1].name)
            assert executed_func_names == {"foo", "bar"}


def test_dry_run(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    target_file = str(absolute_path("_data/loaders/tar/test-archive.tar.gz"))

    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", "general.*", "--dry-run", target_file],
        )

        target_query()
        out, _ = capsys.readouterr()

        assert out == (f"Dry run on: <Target {target_file}>\n  execute: osinfo (general.osinfo.osinfo)\n")


def test_list_json(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test if target-query --list --json output is formatted as we expect it to be."""

    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-query", "-l", "-j"])
        with pytest.raises(SystemExit):
            target_query()
        out, _ = capsys.readouterr()

    try:
        output = json.loads(out)
    except json.JSONDecodeError:
        pytest.fail("Could not decode JSON output of 'target-query --list --json'")

    # test the generic structure of the returned dictionary.
    assert isinstance(output, dict), "Could not load JSON output of 'target-query --list --json'"
    assert output["plugins"], "Expected a dictionary of plugins"
    assert output["loaders"], "Expected a dictionary of loaders"
    assert len(output["plugins"]["loaded"]) > 200, "Expected more loaded plugins"
    assert not output["plugins"].get("failed"), "Some plugin(s) failed to initialize"

    def get_plugin(plugins: list[dict], needle: str) -> dict | bool:
        match = [p for p in plugins["plugins"]["loaded"] if p["name"] == needle]
        return match[0] if match else False

    # general plugin
    users_plugin = get_plugin(output, "users")
    assert users_plugin == {
        "name": "users",
        "description": "Return the users available in the target.",
        "output": "record",
        "arguments": [],
        "alias": False,
        "path": "os.default._os.users",
    }

    # namespaced plugin
    plocate_plugin = get_plugin(output, "plocate.locate")
    assert plocate_plugin == {
        "name": "plocate.locate",
        "description": "Yield file and directory names from the plocate.db.",
        "output": "record",
        "arguments": [],
        "alias": False,
        "path": "os.unix.locate.plocate.locate",
    }

    # regular plugin
    sam_plugin = get_plugin(output, "sam")
    assert sam_plugin == {
        "name": "sam",
        "description": "Dump SAM entries",
        "output": "record",
        "arguments": [],
        "alias": False,
        "path": "os.windows.credential.sam.sam",
    }

    # plugin with arguments
    docker_plugin = get_plugin(output, "docker.logs")
    assert docker_plugin == {
        "name": "docker.logs",
        "output": "record",
        "description": "Returns log files (stdout/stderr) from Docker containers.",
        "arguments": [
            {
                "name": "--raw-messages",
                "type": "bool",
                "help": "preserve ANSI escape sequences and trailing newlines from log messages",
                "default": False,
                "required": False,
            },
            {
                "name": "--remove-backspaces",
                "type": "bool",
                "help": "alter messages by removing ASCII backspaces and the corresponding characters",
                "default": False,
                "required": False,
            },
        ],
        "path": "apps.container.docker.logs",
        "alias": False,
    }


def test_list_json_target_filter(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test if target-query --list "*" --json <target> returns functions we expect it to."""

    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-l=*",
                str(absolute_path("_data/loaders/containerimage/alpine-docker.tar")),
                "-j",
            ],
        )
        with pytest.raises(SystemExit, check=lambda r: r.code == 0):
            target_query()
        out, _ = capsys.readouterr()

    output = json.loads(out)

    # test the generic structure of the returned dictionary.
    assert isinstance(output, dict), "Could not load JSON output of 'target-query --list --json'"
    assert output["plugins"], "Expected a dictionary of plugins"
    assert not output.get("loaders"), "Did not expect a dictionary of loaders"
    assert not output["plugins"].get("failed"), "Some plugin(s) failed to initialize"

    def get_plugin(plugins: list[dict], name: str) -> dict | bool:
        match = [p for p in plugins["plugins"]["loaded"] if p["name"] == name]
        return match[0] if match else False

    # general plugin
    users_plugin = get_plugin(output, "users")
    assert isinstance(users_plugin, dict)

    assert users_plugin == {
        "path": "os.unix.linux._os.users",
        "name": "users",
        "description": "Yield unix user records from passwd files or syslog session logins.",
        "output": "record",
        "arguments": [
            {
                "name": "--sessions",
                "help": "Parse syslog for recent user sessions",
                "default": False,
                "required": False,
                "type": "bool",
            }
        ],
        "alias": False,
    }


def test_record_stream_write_exception_handling(
    caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test if we correctly print the function name of the iterator that failed to iterate."""

    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            ["target-query", "-f", "users,walkfs", str(absolute_path("_data/loaders/tar/test-archive.tar.gz"))],
        )

        with patch("dissect.target.tools.query.record_output", return_value=None):
            target_query()

    assert "Exception occurred while processing output of WalkFSPlugin.walkfs:" in caplog.text


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_arguments_passed_correctly(
    mock_plugins: PluginRegistry, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    class MockPlugin(Plugin):
        def check_compatible(self) -> None:
            pass

        @export(output="none")
        @arg("-j", "--json", action="store_true")
        @arg("--compact", action="store_true")
        def mock_function(self, json: bool, compact: bool) -> None:
            """Mock function to test argument passing."""
            assert json is True
            assert compact is True

    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-fmock",
                str(absolute_path("_data/loaders/tar/test-archive.tar.gz")),
                "--compact",
                "--json",
            ],
        )

        with (
            patch(
                "dissect.target.tools.utils.find_functions",
                autospec=True,
                side_effect=mock_find_functions,
            ),
            patch(
                "dissect.target.target.Target.get_function",
                autospec=True,
                side_effect=lambda self, _: (MockPlugin(self), MockPlugin(self).mock_function),
            ),
        ):
            target_query()

        assert "Exception while executing function mock" not in caplog.text


def test_mixed_namespace_and_regular_regression(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test for regression where executing a namespace and a regular plugin at the same time caused issues."""
    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-query",
                "-f",
                "example_record,example_namespace",
                str(absolute_path("_data/loaders/tar/test-archive.tar.gz")),
                "-s",
            ],
        )

        target_query()
        out, _ = capsys.readouterr()

    assert (
        "<example/descriptor hostname=None domain=None field_a='example' field_b='record'>\n"
        "<example/descriptor hostname=None domain=None field_a='namespace_example' field_b='record'>\n"
    ) in out



from __future__ import annotations

import textwrap
from io import BytesIO, StringIO
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import FileNotFoundError
from dissect.target.helpers.configutil import (
    Bin,
    ConfigurationParser,
    CSVish,
    Default,
    Env,
    Indentation,
    Json,
    Leases,
    ScopeManager,
    SystemD,
    parse,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def parse_data(parser_type: type[ConfigurationParser], data_to_read: str, *args, **kwargs) -> dict:
    """Initializes parser_type as a parser which parses ``data_to_read``."""
    parser = parser_type(*args, **kwargs)
    parser.read_file(StringIO(data_to_read))
    return parser.parsed_data


@pytest.mark.parametrize(
    ("parser_string", "key", "value"),
    [
        ("hello world", "hello", "world"),
        ("hello world\t# new info", "hello", "world"),
    ],
)
def test_unknown_parser(parser_string: str, key: str, value: str) -> None:
    parsed_data = parse_data(Default, parser_string, separator=(r"\s",))
    assert parsed_data[key] == value


@pytest.mark.parametrize(
    ("parser_string", "separator", "expected_output"),
    [
        ("hello=world", "=", {"hello": "world"}),
        ("hello = world", "=", {"hello": "world"}),
        ("hello=world", r"\s", {"hello=world": ""}),
        ("hello-world;20", ";", {"hello-world": "20"}),
        ("hello-world;20;20;20;20", ";", {"hello-world": "20;20;20;20"}),
    ],
)
def test_custom_separators(parser_string: str, separator: tuple, expected_output: dict[str, str]) -> None:
    parsed_data = parse_data(Default, parser_string, separator=separator, comment_prefixes=("#",))
    assert parsed_data == expected_output


@pytest.mark.parametrize(
    ("comment_string", "comment_prefixes"),
    [
        ("; some information", (";", "#")),
        ("# Some other info", (";", "#")),
        ("# Comment ; another comment", (";", "#")),
        ("// code comment", ("//",)),
        (";20;20;20;20", (";",)),
    ],
)
def test_custom_comments(comment_string: str, comment_prefixes: tuple[str, ...]) -> None:
    parsed_data = parse_data(
        Default, f"hello world {comment_string}", separator=(r"\s",), comment_prefixes=comment_prefixes
    )
    assert parsed_data == {"hello": "world"}


@pytest.mark.parametrize(
    ("indented_string", "expected_output"),
    [
        (
            """
            key value1
              value2
            """,
            {"key value1": "value2"},
        ),
        (
            """
            key1 value1
              key2 value2
            """,
            {"key1 value1": {"key2": "value2"}},
        ),
        (
            """
            key value1
              value2
            key value1
              value3
            """,
            {"key value1": ["value2", "value3"]},
        ),
        (
            """
            key value1
              key2 value2a
              key3 value3a
            key value1
              key2 value2b
              key3 value3b
            """,
            {"key value1": [{"key2": "value2a", "key3": "value3a"}, {"key2": "value2b", "key3": "value3b"}]},
        ),
        (
            """
            key value
              key2 value2
              key3 value3

            key value4
              key2 value2
            """,
            {"key value": {"key2": "value2", "key3": "value3"}, "key value4": {"key2": "value2"}},
        ),
        (
            """
            key value
                key2 value2
                key3 value3

            key value4
                key2
            """,
            {"key value": {"key2": "value2", "key3": "value3"}, "key value4": "key2"},
        ),
        (
            """
            key value
                value2
            key2 value
                key3 value3
            """,
            {"key value": "value2", "key2 value": {"key3": "value3"}},
        ),
    ],
)
def test_indented_parser(indented_string: str, expected_output: dict[str, dict]) -> None:
    parsed_data = parse_data(Indentation, textwrap.dedent(indented_string), separator=(r"\s",))
    assert parsed_data == expected_output


def test_collapse_inversion() -> None:
    input_data = "hello world\nhello test\nworld world\nworld test"
    assert parse_data(
        Default,
        input_data,
        separator=(r"\s",),
        collapse={"hello"},
        collapse_inverse=True,
    ) == {"hello": ["world", "test"], "world": "test"}


def test_change_scope() -> None:
    parser = Indentation(separator=(r"\s",))
    manager = ScopeManager()

    # Scoping does not change
    changed = parser._change_scope(manager, line="key value", next_line="key2 value2", key="key value")
    assert id(manager._root) == id(manager._current)
    assert not changed
    old_current = manager._current

    # Scoping changes once
    changed = parser._change_scope(manager, line="key2 value2", next_line="  value2", key="key2 value2")
    assert id(old_current) != id(manager._current)
    assert changed
    assert manager._root == {"key2 value2": {}}
    old_current = manager._current

    # If the current line still contains empty space, return the same scope
    changed = parser._change_scope(manager, line="  value2", next_line="test_line", key="value2")
    assert id(old_current) == id(manager._current)
    assert not changed

    changed = parser._change_scope(manager, line="test data", next_line=None, key="test data")
    assert id(old_current) == id(manager._current)
    assert not changed


@pytest.mark.parametrize(
    ("string_data", "expected_output"),
    [
        ("[Unit]\n[System]\n", {"Unit": {}, "System": {}}),
        ("[Unit]\nkey=value;\n[System]", {"Unit": {"key": "value;"}, "System": {}}),
        (
            "[Unit]\nkey=value \\\n continued\\\nvalue\nkey2=value2",
            {
                "Unit": {
                    "key": "value continued value",
                    "key2": "value2",
                }
            },
        ),
        (
            "[Unit]\nkey=value \\\n continued\\\n[System]",
            {
                "Unit": {
                    "key": "value continued",
                },
                "System": {},
            },
        ),
        (
            "[Unit]\nnew_lines=hello \\\nworld\\\ntest\\\n help",
            {
                "Unit": {"new_lines": "hello world test help"},
            },
        ),
    ],
    ids=["scoping changes", "key value extraction", "line continuation", "faulty configuration", "weird indentation"],
)
def test_systemd_scope(string_data: str, expected_output: dict) -> None:
    parser = SystemD()

    parser.parse_file(StringIO(string_data))

    assert parser.parsed_data == expected_output


def test_systemd_basic_syntax() -> None:
    data = Path(absolute_path("_data/helpers/configutil/systemd.syntax"))
    output = {
        "Section A": {
            "KeyOne": "value 1",
            "KeyTwo": "value 2",
        },
        "Section B": {
            "Setting": '"something" "some thing" "..."',
            "KeyTwo": "value 2 value 2 continued",
        },
        "Section C": {
            "KeyThree": "value 3 value 3 continued",
        },
    }

    parser = SystemD()
    parser.parse_file(StringIO(data.read_text()))

    assert parser.parsed_data == output


@pytest.mark.parametrize(
    ("data_string", "expected_data"),
    [
        (r'{"data" : "value"}', {"data": "value"}),
        (r'[{"data" : "value"}]', {"list_item0": {"data": "value"}}),
        (
            r'[{"data" : "value"}, {"data" : "value2"}]',
            {"list_item0": {"data": "value"}, "list_item1": {"data": "value2"}},
        ),
        (
            r'[{"data": [{"key1": "value1"}, {"key1": "value2"}]}]',
            {
                "list_item0": {
                    "data": {
                        "list_item0": {"key1": "value1"},
                        "list_item1": {"key1": "value2"},
                    },
                },
            },
        ),
    ],
)
def test_json_syntax(data_string: str, expected_data: dict | list) -> None:
    parser = Json()
    parser.parse_file(StringIO(data_string))

    assert parser.parsed_data == expected_data


@pytest.mark.parametrize(
    ("data", "expected_data"),
    [
        (b"\x00\x01\x02", {"binary": b"\x00\x01\x02", "size": "3"}),
    ],
)
def test_bin_parser(data: bytes, expected_data: dict) -> None:
    parser = Bin()
    parser.parse_file(BytesIO(data))
    assert parser.parsed_data == expected_data


@pytest.mark.parametrize(
    ("fields", "separator", "comment", "data_string", "expected_data"),
    [
        (["a", "b"], (r"\s+",), ("#",), "1 2", {"0": {"a": "1", "b": "2"}}),  # SSV
        (["a", "b"], (r"\s+",), ("#",), "1 2\n3 4", {"0": {"a": "1", "b": "2"}, "1": {"a": "3", "b": "4"}}),  # SSV-2
        (["a", "b"], (r"\s+",), ("#",), "1\t2", {"0": {"a": "1", "b": "2"}}),  # TSV
        (["a", "b"], (r",",), ("#",), "1,2", {"0": {"a": "1", "b": "2"}}),  # CSV
        (["a", "b"], (r"\|",), ("#",), "1|2", {"0": {"a": "1", "b": "2"}}),  # DSV
        (["a", "b"], (r"\|",), ("#",), "#9|9\n1|2", {"0": {"a": "1", "b": "2"}}),  # comments
        (["a", "b"], (r"\s+",), ("#",), "#9 9\n1 2 3", {"0": {"a": "1", "b": "2 3"}}),  # trailing column
        (["a", "b"], (r"\s+",), ("#",), "x", {"0": {"line": "x"}}),  # unparsed
    ],
)
def test_csv_syntax(
    fields: list[str], separator: tuple[str, ...], comment: tuple[str, ...], data_string: str, expected_data: dict
) -> None:
    parser = CSVish(fields=fields, separator=separator, comment_prefixes=comment)
    parser.parse_file(StringIO(data_string))
    assert parser.parsed_data == expected_data


def test_parse(target_linux: Target, fs_linux: VirtualFilesystem, tmp_path: Path) -> None:
    # File does not exist on the system in the first place
    with pytest.raises(FileNotFoundError):
        parse(target_linux.fs.path("/path/to/file"))

    file_path = tmp_path.joinpath("path/to/file")
    file_path.parent.mkdir(parents=True)
    file_path.touch()

    fs_linux.map_dir("/", tmp_path.absolute())

    # Trying to read a directory
    with pytest.raises(FileNotFoundError):
        parse(target_linux.fs.path("/path/to"))

    parse(target_linux.fs.path("/path/to/file"))


@pytest.mark.parametrize(
    ("input", "expected_output"),
    [
        ("", {}),
        ("foo=bar\n", {"foo": "bar"}),
        ("foo = bar \n", {"foo": "bar"}),
        ('foo = "bar"\n', {"foo": "bar"}),
        ("foo=bar\nempty=\nhello=world\n", {"foo": "bar", "empty": "", "hello": "world"}),
        # multiple delimiters
        ("foo=option=foo,value=bar", {"foo": "option=foo,value=bar"}),
        # types
        ("foo=123", {"foo": 123}),
        ("foo=1", {"foo": True}),
        ("foo=0", {"foo": False}),
        ("foo='1'", {"foo": True}),
        ("foo=true", {"foo": True}),
        ('foo="False"\nbar="true"\n', {"bar": True, "foo": False}),
        # comments
        ("foo=bar#\n", {"foo": "bar#"}),
        ('foo="bar#"\n', {"foo": "bar#"}),
        ("foo='bar#'\n", {"foo": "bar#"}),
        ("foo=bar # comment\n", {"foo": "bar"}),
        ("foo='bar' # comment\n", {"foo": "bar"}),
        ("foo='bar #' # comment\n", {"foo": "bar #"}),
        ("foo=bar# not a comment\n", {"foo": "bar# not a comment"}),
        ("foo='bar # not a comment'\n", {"foo": "bar # not a comment"}),
        (Path(absolute_path("_data/helpers/configutil/env")), {"foo": "'bar"}),
        # whitespaces inside quotes
        ("foo='bar '\n", {"foo": "bar "}),
        # quotes inside values
        ('foo=ba"r\n', {"foo": 'ba"r'}),
    ],
)
def test_env_parser(input: str | Path, expected_output: dict) -> None:
    parser = Env(comments=False)

    if isinstance(input, str):
        parser.parse_file(StringIO(input))
    else:
        parser.parse_file(input.open("r"))

    assert parser.parsed_data == expected_output


@pytest.mark.parametrize(
    ("string_data", "expected_output"),
    [
        (
            'default-duid "\\000\\001\\000\\001\\037\\305\\371\\341\\001\\002\\003\\004\\005\\006"\nlease {\ninterface "eth0"; # some comment\nfixed-address "1.2.3.4";\noption subnet-mask 255.255.255.0;\nrenew 2 2023/10/04 13:37:04;\n# some other comment}',  # noqa: E501
            {
                "default-duid": '"\\000\\001\\000\\001\\037\\305\\371\\341\\001\\002\\003\\004\\005\\006"',
                "lease-0": {
                    "interface": '"eth0"',
                    "fixed-address": '"1.2.3.4"',
                    "option": {"subnet-mask": "255.255.255.0"},
                    "renew": "2 2023/10/04 13:37:04",
                },
            },
        ),
    ],
)
def test_leases_parser(string_data: str, expected_output: dict) -> None:
    """Test the leases parser."""
    parser = Leases()
    parser.parse_file(StringIO(string_data))

    assert parser.parsed_data == expected_output



from __future__ import annotations

import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.evidence import ad1

from dissect.target.exceptions import (
    FileNotFoundError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class AD1Filesystem(Filesystem):
    __type__ = "ad1"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.ad1 = ad1.AD1(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return fh.read(16) == b"ADSEGMENTEDFILE\x00"

    def get(self, path: str) -> AD1FilesystemEntry:
        return AD1FilesystemEntry(self, path, self._get_entry(path))

    def _get_entry(self, path: str) -> ad1.LogicalImage | ad1.FileEntry:
        try:
            return self.ad1.get(path)
        except IOError:
            raise FileNotFoundError(path)


class AD1FilesystemEntry(FilesystemEntry):
    def get(self, path: str) -> AD1FilesystemEntry:
        path = fsutil.join(self.path, path, alt_separator=self.alt_separator)
        return AD1FilesystemEntry(self.fs, path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self.entry.open()

    def iterdir(self) -> Iterator[str]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        yield from self.entry.listdir().keys()

    def scandir(self) -> Iterator[AD1FilesystemEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for fname, file_ in self.entry.listdir().items():
            path = fsutil.join(self.path, fname, alt_separator=self.alt_separator)
            yield AD1FilesystemEntry(self.fs, path, file_)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return self.entry.is_file()

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        return self.entry.is_dir()

    def is_symlink(self) -> bool:
        return False

    def readlink(self) -> str:
        raise NotASymlinkError

    def readlink_ext(self) -> FilesystemEntry:
        raise NotASymlinkError

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        size = self.entry.size if self.entry.is_file() else 0
        entry_addr = fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator)
        return fsutil.stat_result([stat.S_IFREG, entry_addr, id(self.fs), 1, 0, 0, size, 0, 0, 0])



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from dissect.sql import SQLite3
from dissect.util.ts import wintimestamp
from flow.record.fieldtypes import digest, windows_path

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import (
    TargetRecordDescriptor,
    create_extended_descriptor,
)
from dissect.target.helpers.regutil import (
    RegistryKey,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
)
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.sql.sqlite3 import Row

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


CamRegistryRecord = create_extended_descriptor([RegistryRecordDescriptorExtension, UserRecordDescriptorExtension])(
    "windows/cam/registry",
    [
        ("datetime", "ts"),
        ("string", "device"),
        ("string", "app_name"),
        ("path", "path"),
        ("datetime", "last_started"),
        ("datetime", "last_stopped"),
        ("varint", "duration"),
    ],
)

CamUsageHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/cam/usagehistory",
    [
        ("datetime", "last_used_time_stop"),
        ("datetime", "last_used_time_start"),
        ("varint", "duration"),
        ("string", "package_type"),
        ("string", "capability"),
        ("string", "file_id"),
        ("digest", "file_id_hash"),
        ("string", "access_blocked"),
        ("string", "program_id"),
        ("string", "package_family_name"),
        ("string", "access_guid"),
        ("string", "label"),
        ("string", "app_name"),
        ("path", "binary_full_path"),
        ("string", "service_name"),
    ],
)

CamIdentityRelationshipHistoryRecord = TargetRecordDescriptor(
    "windows/cam/identityrelationshiphistory",
    [
        ("datetime", "last_observed_time"),
        ("string", "package_type"),
        ("string", "file_id"),
        ("digest", "file_id_hash"),
        ("string", "program_id"),
        ("path", "binary_full_path"),
    ],
)

CamGlobalPromptHistoryRecord = TargetRecordDescriptor(
    "windows/cam/globalprompthistory",
    [
        ("datetime", "shown_time"),
        ("string", "package_type"),
        ("string", "capability"),
        ("string", "file_id"),
        ("digest", "file_id_hash"),
        ("string", "program_id"),
    ],
)


class CamPlugin(Plugin):
    """Plugin that iterates various Capability Access Manager registry key locations."""

    __namespace__ = "cam"

    CONSENT_STORES = (
        "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\CapabilityAccessManager\\ConsentStore",
        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\CapabilityAccessManager\\ConsentStore",
    )

    CAP_DB_REG_PATH = (
        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\CapabilityAccessManager\\CapabilityUsageHistory"
    )

    CONTENT_TABLES = (
        "NonPackagedUsageHistory",
        "PackagedUsageHistory",
        "NonPackagedIdentityRelationship",
        "NonPackagedGlobalPromptHistory",
    )
    CONTEXT_TABLES = (
        "Capabilities",
        "PackageFamilyNames",
        "BinaryFullPaths",
        "Users",
        "FileIDs",
        "ProgramIDs",
        "AccessGUIDs",
        "AppNames",
        "ServiceNames",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.app_regf_keys = self._find_apps()
        self.camdb_path = self._find_db()
        self.camdb: SQLite3 | None = None

    def _find_apps(self) -> list[RegistryKey]:
        return [key for store in self.target.registry.keys(self.CONSENT_STORES) for key in store.subkeys()]

    def check_compatible(self) -> None:
        if not self.app_regf_keys and not self.camdb_path:
            raise UnsupportedPluginError("No Capability Access Manager keys found")

    def yield_apps(self) -> Iterator[RegistryKey]:
        for app in self.app_regf_keys:
            for key in app.subkeys():
                if key.name == "NonPackaged":  # NonPackaged registry key has more apps, so yield those apps
                    yield from key.subkeys()
                else:
                    yield key

    def _find_db(self) -> TargetPath | None:
        try:
            # Retrieve the location of the CapabilityAccessManager.db directory from registry
            DatabaseRoot = self.target.registry.key(self.CAP_DB_REG_PATH).value("DatabaseRoot").value
            camdb_path = self.target.fs.path(DatabaseRoot).joinpath("CapabilityAccessManager.db")

            if camdb_path.exists():
                return camdb_path
        except RegistryKeyNotFoundError:
            # This database and registry key only exists on recent Windows 11 systems.
            self.target.log.warning(
                "cam.history: Cannot find database location registry key, OS probably not supported"
            )

    def _open_db(self) -> SQLite3 | None:
        return SQLite3(self.camdb_path.open("rb")) if self.camdb_path else None

    def _build_context_dict(self) -> defaultdict[str, dict] | None:
        MAPDB = defaultdict(dict)
        for table_name in self.CONTEXT_TABLES:
            for row in self._yield_rows(table_name):
                MAPDB[table_name][row.get("ID")] = row.get("StringValue")
        return MAPDB

    def _convert_ts(self, ts: int) -> None:
        if not ts:
            return None
        return wintimestamp(ts)

    def _convert_file_id(self, file_id: str) -> None:
        return digest((None, file_id[4:], None)) if file_id else None

    def _yield_rows(self, table_name: str) -> Iterator[Row]:
        if not (table := self.camdb.table(table_name)):
            return
        yield from table

    @export(record=[CamUsageHistoryRecord, CamIdentityRelationshipHistoryRecord, CamGlobalPromptHistoryRecord])
    def history(
        self,
    ) -> Iterator[CamUsageHistoryRecord | CamIdentityRelationshipHistoryRecord | CamGlobalPromptHistoryRecord]:
        """Iterate Capability Access Manager History entries.
        The Capability Access Manager keeps track of processes that access I/O devices, like the webcam or microphone.
        Applications are divided into packaged and non-packaged applications meaning Microsoft or
        non-Microsoft applications. Additional historical entries are since Windows 11 available in a SQL database.

        Records are created from the following tables:
        - NonPackagedUsageHistory
        - PackagedUsageHistory
        - NonPackagedIdentityRelationship
        - NonPackagedGlobalPromptHistory

        References:
            - https://medium.com/@cyber.sundae.dfir/capability-access-manager-forensics-in-windows-11-f586ef8aac79

        Yields ``CamUsageHistoryRecord``, ``CamIdentityRelationshipHistoryRecord`` or ``CamGlobalPromptHistoryRecord``:

        Record CamUsageHistoryRecord:

        .. code-block:: text

            last_used_time_stop (datetime): When the application last stopped using the capability.
            last_used_time_start (datetime): When the application last started using the capability.
            duration (varint): How long the application used the capability.
            package_type (string): The application type of the record, originates from the table name.
            capability (string): The capability being used; microphone, camera, location etc.
            file_id (string): The sha1 hash of the application making use of the capability.
            file_id_hash (digest): Digest version of the file_id field.
            access_blocked (string): If capability usage was allowed, 0 = Not blocked and 1 = blocked.
            program_id (string): Program ID of application, unclear what this value means.
            package_family_name (string): Package name of application using capability.
            access_guid (string): Unclear what the value of this is.
            label (string): Unclear what the value of this is, no joinable table with this ID.
            app_name (string): Name of the application using capability.
            binary_full_path (path): Path of the application using capability.
            service_name (string): Name of the service using capability.

        Record CamIdentityRelationshipHistoryRecord:

        .. code-block:: text

            last_observed_time (datetime): Last time capability was observed.
            package_type (string): The application type of the record, originates from the table name.
            file_id (string): The sha1 hash of the application making use of the capability.
            file_id_hash (digest): Digest version of the file_id field.
            program_id (string): Program ID of application, unclear what this value means.
            binary_full_path (path): Path of the application using capability.

        Record CamGlobalPromptHistoryRecord

        .. code-block:: text

            shown_time (datetime): Last time capability was observed.
            package_type (string): The application type of the record, originates from the table name.
            capability (string): The capability being used; microphone, camera, location etc.
            file_id (string): The sha1 hash of the application making use of the capability.
            file_id_hash (digest): Digest version of the file_id field.
            program_id (string): Program ID of application, unclear what this value means.
        """

        self.camdb = self._open_db()

        # Silently exit the function if no database object could be created.
        # This means the function is not compatible with the target.
        if not self.camdb:
            return None

        # Create mapping dict for sql 'join' actions.
        MAPDB = self._build_context_dict()

        # Process NonPackagedUsageHistory and PackagedUsageHistory tables
        for table_name in ["NonPackagedUsageHistory", "PackagedUsageHistory"]:
            for row in self._yield_rows(table_name):
                # Resolve user from UserSID
                if user := self.target.user_details.find(MAPDB["Users"].get(row["UserSid"], "")):
                    user = user.user

                # Calculate duration field
                duration = 0
                last_used_time_stop = self._convert_ts(row.get("LastUsedTimeStop"))
                last_used_time_start = self._convert_ts(row.get("LastUsedTimeStart"))

                if all([last_used_time_stop, last_used_time_start]):
                    duration = last_used_time_stop.timestamp() - last_used_time_start.timestamp()

                yield CamUsageHistoryRecord(
                    last_used_time_stop=last_used_time_stop,
                    last_used_time_start=last_used_time_start,
                    duration=duration,
                    package_type=table_name,
                    capability=MAPDB["Capabilities"].get(row["Capability"]),
                    file_id=MAPDB["FileIDs"].get(row["FileID"]),
                    file_id_hash=self._convert_file_id(MAPDB["FileIDs"].get(row["FileID"])),
                    access_blocked=row["AccessBlocked"],
                    access_guid=MAPDB["AccessGUIDs"].get(row["AccessGUID"]),
                    label=row["Label"],
                    app_name=MAPDB["AppNames"].get(row["AppName"]),
                    program_id=MAPDB["ProgramIDs"].get(row["ProgramID"]),
                    binary_full_path=MAPDB["BinaryFullPaths"].get(row["BinaryFullPath"]),
                    package_family_name=MAPDB["PackageFamilyNames"].get(row["PackageFamilyName"]),
                    service_name=MAPDB["ServiceNames"].get(row["ServiceName"]),
                    _target=self.target,
                    _user=user,
                )

        # Process table NonPackagedIdentityRelationship
        for row in self._yield_rows("NonPackagedIdentityRelationship"):
            yield CamIdentityRelationshipHistoryRecord(
                last_observed_time=self._convert_ts(row.get("LastObservedTime")),
                package_type=table_name,
                file_id=MAPDB["FileIDs"].get(row["FileID"]),
                file_id_hash=self._convert_file_id(MAPDB["FileIDs"].get(row["FileID"])),
                program_id=MAPDB["ProgramIDs"].get(row["ProgramID"]),
                binary_full_path=MAPDB["BinaryFullPaths"].get(row["BinaryFullPath"]),
                _target=self.target,
            )

        # Process table NonPackagedGlobalPromptHistory
        for row in self._yield_rows("NonPackagedGlobalPromptHistory"):
            yield CamGlobalPromptHistoryRecord(
                shown_time=self._convert_ts(row.get("ShownTime")),
                package_type=table_name,
                capability=MAPDB["Capabilities"].get(row["Capability"]),
                file_id=MAPDB["FileIDs"].get(row["FileID"]),
                file_id_hash=self._convert_file_id(MAPDB["FileIDs"].get(row["FileID"])),
                program_id=MAPDB["ProgramIDs"].get(row["ProgramID"]),
                _target=self.target,
            )

    @export(record=CamRegistryRecord)
    def registry(self) -> Iterator[CamRegistryRecord]:
        """Iterate Capability Access Manager key locations.

        The Capability Access Manager keeps track of processes that access I/O devices, like the webcam or microphone.
        Applications are divided into packaged and non-packaged applications meaning Microsoft or
        non-Microsoft applications.

        References:
            - https://docs.velociraptor.app/exchange/artifacts/pages/windows.registry.capabilityaccessmanager/
            - https://svch0st.medium.com/can-you-track-processes-accessing-the-camera-and-microphone-7e6885b37072

        Yields ``CamRegistryRecord`` with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The modification timestamp of the registry key.
            device (string): Name of the device privacy permission where asked for.
            app_name (string): The name of the application.
            path (path): The possible path to the application.
            last_started (datetime): When the application last started using the device.
            last_stopped (datetime): When the application last stopped using the device.
            duration (varint): How long the application used the device (seconds).
        """

        for key in self.yield_apps():
            last_started = None
            last_stopped = None
            duration = None

            try:
                last_started = wintimestamp(key.value("LastUsedTimeStart").value)
            except RegistryValueNotFoundError:
                self.target.log.warning("No LastUsedTimeStart for application: %s", key.name)

            try:
                last_stopped = wintimestamp(key.value("LastUsedTimeStop").value)
            except RegistryValueNotFoundError:
                self.target.log.warning("No LastUsedTimeStop for application: %s", key.name)

            if last_started and last_stopped:
                duration = (last_stopped - last_started).seconds

            # The "device" is derived from the subkey (webcam in this case) of ConsentStore example:
            # HKLM\\...\\CapabilityAccessManager\\ConsentStore\\webcam\\C:#Program Files#Mozilla Firefox#firefox.exe
            # However, sometimes there is an additional subfolder such as "NonPackaged" in between. Example:
            # HKLM\\...\\CapabilityAccessManager\\ConsentStore\\webcam\\NonPackaged\\C:#...#Mozilla Firefox#firefox.exe
            # Code below determines the index of 'ConsentStore' (always present) in the path and grabs the sub-folder.
            path_list = key.path.split("\\")
            device = path_list[path_list.index("ConsentStore") + 1]

            # The application/or path of the application using the device is the key.name itself:
            # C:#Program Files#Mozilla Firefox#firefox.exe
            path = windows_path(key.name.replace("#", "\\")) if "#" in key.name else None

            yield CamRegistryRecord(
                ts=key.ts,
                device=device,
                app_name=key.name,
                path=path,
                last_started=last_started,
                last_stopped=last_stopped,
                duration=duration,
                _target=self.target,
                _key=key,
                _user=self.target.registry.get_user(key),
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import (
    TargetRecordDescriptor,
    create_extended_descriptor,
)
from dissect.target.plugin import Plugin, arg, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

ExampleRecordRecord = TargetRecordDescriptor(
    "example/descriptor",
    [
        ("string", "field_a"),
        ("string", "field_b"),
    ],
)

ExampleUserRegistryRecord = create_extended_descriptor(
    [RegistryRecordDescriptorExtension, UserRecordDescriptorExtension]
)(
    "example/registry/user",
    [
        ("datetime", "ts"),
    ],
)


class ExamplePlugin(Plugin):
    """Example plugin.

    This plugin serves as an example for new plugins. Use it as a guideline.

    Docstrings are used in help messages of plugins. Make sure to document
    your plugin and plugin functions. Use Google docstring format:

    https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html

    Plugins can optionally be namespaced by specifying the ``__namespace__``
    class attribute. Namespacing results in your plugin needing to be prefixed
    with this namespace when being called. For example, if your plugin has
    specified ``test`` as namespace and a function called ``example``, you must
    call your plugin with ``test.example``::

        __namespace__ = "test"

    The ``__init__`` takes the target as only argument. Perform additional
    initialization here if necessary::

        def __init__(self, target: Target):
            super().__init__(target)
    """

    # IMPORTANT: Remove these attributes when using this as boilerplate for your own plugin!
    __findable__ = False

    def check_compatible(self) -> None:
        """Perform a compatibility check with the target.

        This function should return ``None`` if the plugin is compatible with
        the current target (``self.target``). For example, check if a certain
        file exists.
        Otherwise it should raise an ``UnsupportedPluginError``.

        Raises:
            UnsupportedPluginError: If the plugin could not be loaded.
        """

    @export(output="default")
    @arg("--flag", action="store_true", help="optional example flag")
    def example(self, flag: bool = False) -> str:
        """Example plugin function.

        Docstrings are used in help messages of plugins. Make sure to document
        your plugin and plugin functions. The first line must be a brief one
        sentence description of the plugin function.

        The ``@export`` decorator supports multiple arguments:
            property (bool): Whether this function should act like a property.
                Properties are implicitly cached.
            record (RecordDescriptor): The record descriptor this function yield,
                if any. If dynamic, use :class:`~dissect.target.helpers.record.DynamicDescriptor`.
            output (str): The output type of this function. Can be one of:
                - default: Single return value.
                - record: Yields records. Implicit when record argument is given.
                - yield: Yields printable values.
                - none: No return value.

        Command line arguments can be added using the ``@arg`` decorator. Arguments
        to this decorator are directly forwarded to the ``add_argument`` function
        of `argparse <https://docs.python.org/library/argparse.html>`_.
        Resulting arguments are passed to the function using kwargs.
        The keyword argument name must match the argparse argument name.
        """
        return f"Example plugin. Flag argument: {flag!r}"

    @export(record=ExampleRecordRecord)
    def example_record(self) -> Iterator[ExampleRecordRecord]:
        """Example plugin that generates records.

        To create a new plugin function that yields records, you must define a record descriptor
        and pass it to ``@export``. This will implicitly mark the output type as ``record``.
        """
        yield ExampleRecordRecord(
            field_a="example",
            field_b="record",
            _target=self.target,
        )

    @export(record=ExampleUserRegistryRecord)
    def example_user_registry_record(self) -> Iterator[ExampleUserRegistryRecord]:
        """Example plugin that generates records with registry key and user information.

        To include registry or user information in a record, you must create a new record descriptor using
        :func:`~dissect.target.helpers.record.create_extended_descriptor` with
        :class:`~dissect.target.helpers.descriptor_extensions.RegistryRecordDescriptorExtension` and/or
        :class:`~dissect.target.helpers.descriptor_extensions.UserRecordDescriptorExtension` as extensions.
        """
        for key in self.target.registry.keys("HKCU\\SOFTWARE"):
            user = self.target.registry.get_user(key)
            yield ExampleUserRegistryRecord(
                ts=key.ts,
                _key=key,
                _user=user,
                _target=self.target,
            )

    @export(output="yield")
    def example_yield(self) -> Iterator[str]:
        """Example plugin that yields text lines.

        Setting ``output="yield"`` is useful for creating generators of text, such as human-readable timelines.
        """
        for i in range(10):
            yield f"Example line {i}"

    @export(output="none")
    def example_none(self) -> None:
        """Example plugin with no return value.

        Setting ``output="none"`` means you don't return a value. This is useful when you want to print something
        on your own, such as verbose information.
        """
        print("Example output with no return value.")

    @internal
    def example_internal(self) -> str:
        """Example internal plugin.

        Use the ``@internal`` plugin to mark your plugin as internal and hide it from the plugin overview.
        """
        return "Example internal plugin."


class ExampleNamespacePlugin(Plugin):
    """Example namespace plugin."""

    __namespace__ = "example_namespace"
    __findable__ = False

    def check_compatible(self) -> None:
        pass

    @export(record=ExampleRecordRecord)
    def example_record(self) -> Iterator[ExampleRecordRecord]:
        """Example namespace export."""
        yield ExampleRecordRecord(
            field_a="namespace_example",
            field_b="record",
            _target=self.target,
        )



from __future__ import annotations

import csv
import io
import ipaddress
import struct
from pathlib import Path
from typing import TYPE_CHECKING, Final

from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

SEPLogRecord = TargetRecordDescriptor(
    "application/av/symantec/sep/log",
    [
        ("datetime", "ts"),
        ("string", "virus"),
        ("string", "user"),
        ("path", "source_file"),
        ("string", "action_taken"),
        ("string", "virus_type"),
        ("string", "scan_id"),
        ("string", "quarantine_id"),
        ("varint", "virus_id"),
        ("varint", "depth"),
        ("boolean", "still_infected"),
        ("boolean", "quarantined"),
        ("boolean", "compressed"),
        ("boolean", "cleanable"),
        ("boolean", "deletable"),
        ("varint", "confidence"),
        ("varint", "prevalence"),
        ("varint", "risk"),
        ("uri", "download_url"),
        ("varint", "line_no"),
    ],
)

SEPFirewallRecord = TargetRecordDescriptor(
    "application/av/symantec/sep/firewall",
    [
        ("datetime", "ts"),
        ("string", "protocol"),
        ("net.ipaddress", "local_ip"),
        ("net.ipaddress", "remote_ip"),
        ("net.ipaddress", "local_ip6"),
        ("net.ipaddress", "remote_ip6"),
        ("varint", "local_port"),
        ("varint", "remote_port"),
        ("boolean", "outbound"),
        ("datetime", "begin_time"),
        ("datetime", "end_time"),
        ("varint", "repetition"),
        ("boolean", "blocked"),
        ("string", "severity"),
        ("varint", "rule_id"),
        ("string", "remote_host"),
        ("string", "rule_name"),
        ("path", "application"),
        ("string", "user"),
        ("varint", "line_no"),
    ],
)


class SymantecPlugin(Plugin):
    """Symantec Endpoint Security Suite Plugin.

    References:
        - https://malwaremaloney.blogspot.com/2021/01/
    """

    __namespace__ = "symantec"

    LOG_SEP_AV = "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/Logs/AV/*"
    LOG_SEP_NET = "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/Logs/tralog.log"

    LOGS = (LOG_SEP_AV, LOG_SEP_NET)

    # Special values
    MARKER_INFECTION = 5
    QUARANTINE_SUCCESS = 2
    CLEANABLE = 0
    DELETABLE = 4
    STILL_INFECTED = 1
    OUTBOUND = 2
    BLOCKED = 1
    COMPRESSED = 1

    TCP_INIT = 301
    TCP_CLOSE = 304
    UDP_DATA = 302

    # Log fields
    AV_TIMESTAMP = 0
    AV_EVENT = 1
    AV_USER = 5
    AV_VIRUS = 6
    AV_FILE = 7
    AV_ACTION_TAKEN = 10
    AV_VIRUS_TYPE = 11
    AV_SCAN_ID = 14
    AV_EVENT_DATA = 17
    AV_QUARANTINE_ID = 18
    AV_VIRUS_ID = 19
    AV_QUARANTINE_STATUS = 20
    AV_COMPRESSED = 23
    AV_DEPTH = 24
    AV_STILL_INFECTED = 25
    AV_CLEANABLE = 28
    AV_DELETABLE = 29
    AV_CONFIDENCE = 65
    AV_PREVALENCE = 67
    AV_DOWNLOADED_FROM = 68
    AV_RISK = 71

    # Firwall fields
    FW_TIMESTAMP = 1
    FW_PROTOCOL = 2
    FW_LOCAL_IP = 3
    FW_REMOTE_IP = 4
    FW_LOCAL_PORT = 5
    FW_REMOTE_PORT = 6
    FW_DIRECTION = 7
    FW_BEGIN_TIME = 8
    FW_END_TIME = 9
    FW_REPETITION = 10
    FW_ACTION = 11
    FW_SEVERITY = 13
    FW_RULE_ID = 14
    FW_REMOTE_HOST_NAME = 15
    FW_RULE_NAME = 16
    FW_APPLICATION = 17
    FW_LOCATION = 20
    FW_USER = 21
    FW_LOCAL_IP6 = 25
    FW_REMOTE_IP6 = 26

    PROTOCOL: Final[dict[int, str]] = {
        301: "TCP initiated",
        302: "UDP datagram",
        303: "Ping request",
        304: "TCP completed",
        305: "Traffic (other)",
        306: "ICMPv4 packet",
        307: "Ethernet packet",
        308: "IP packet",
        309: "ICMPv6 packet",
    }

    SEVERITY = ["Critical"] * 4 + ["Major"] * 4 + ["Minor"] * 4 + ["Info"] * 4

    ACTION: Final[dict[int, str]] = {
        1: "Quarantine",
        2: "Rename",
        3: "Delete",
        4: "Leave Alone",
        5: "Clean",
        6: "Remove Macros",
        7: "Save file as...",
        8: "Sent to backend",
        9: "Restore from Quarantine",
        11: "Undo Action",
        12: "Error",
        13: "Backup to quarantine (backup view)",
        14: "Pending Analysis",
        15: "Partially Fixed",
        16: "Terminate Process Required",
        17: "Exclude from Scanning",
        18: "Reboot Processing",
        19: "Clean by Deletion",
        20: "Access Denied",
        21: "TERMINATE PROCESS ONLY",
        22: "NO REPAIR",
        23: "FAIL",
        24: "RUN POWERTOOL",
        25: "NO REPAIR POWERTOOL",
        110: "INTERESTING PROCESS CAL",
        111: "INTERESTING PROCESS DETECTED",
        1000: "INTERESTING PROCESS HASHED DETECTED",
        1001: "DNS HOST FILE EXCEPTION",
    }

    VIRUS_TYPE: Final[dict[int, str]] = {
        48: "Heuristic",
        64: "Reputation",
        80: "Hack Tools",
        96: "Spyware",
        112: "Trackware",
        128: "Dialers",
        144: "Remote Access",
        160: "Adware",
        176: "Joke Programs",
        224: "Heuristic Application",
        256: "Test",
    }

    def __init__(self, target: Target):
        super().__init__(target)
        self.codepage = self.target.codepage or "ascii"

    def check_compatible(self) -> None:
        for log_file in self.LOGS:
            if list(self.target.fs.glob(log_file)):
                return
        raise UnsupportedPluginError("No Symantec SEP logs found")

    def _fw_cell(self, line: list, cell_id: int) -> str:
        return line[cell_id].decode("utf-8")

    def _fw_hex_cell(self, line: list, cell_id: int) -> int:
        return int(self._fw_cell(line, cell_id), 16)

    def _septime(self, ts: str) -> str:
        # Date/Time stored as 12 digit hex number (6 pairs of 2 digits)
        # Offsets for year (+1970) and month (+1), no TZ
        year = int(ts[0:2], 16)
        month = int(ts[2:4], 16)
        day = int(ts[4:6], 16)
        hour = int(ts[6:8], 16)
        minute = int(ts[8:10], 16)
        second = int(ts[10:12], 16)
        return f"{year + 1970}-{month + 1:0>2}-{day:0>2}T{hour:0>2}:{minute:0>2}:{second:0>2}"

    def _fw_line(self, line: bytes, line_no: int) -> Iterator[SEPFirewallRecord]:
        try:
            cells = line.split(b"\t")
            protocol_id = self._fw_hex_cell(cells, self.FW_PROTOCOL)
            tcp_udp = protocol_id in [self.TCP_INIT, self.TCP_CLOSE, self.UDP_DATA]

            yield SEPFirewallRecord(
                ts=wintimestamp(self._fw_hex_cell(cells, self.FW_TIMESTAMP)),
                protocol=self.PROTOCOL.get(protocol_id, "Unknown"),
                local_ip=ipaddress.ip_address(struct.pack("<i", self._fw_hex_cell(cells, self.FW_LOCAL_IP))),
                remote_ip=ipaddress.ip_address(struct.pack("<i", self._fw_hex_cell(cells, self.FW_REMOTE_IP))),
                local_ip6=ipaddress.ip_address(ipaddress.v6_int_to_packed(self._fw_hex_cell(cells, self.FW_LOCAL_IP6))),
                remote_ip6=ipaddress.ip_address(
                    ipaddress.v6_int_to_packed(self._fw_hex_cell(cells, self.FW_REMOTE_IP6))
                ),
                local_port=self._fw_hex_cell(cells, self.FW_LOCAL_PORT) if tcp_udp else None,
                remote_port=self._fw_hex_cell(cells, self.FW_REMOTE_PORT) if tcp_udp else None,
                outbound=self._fw_hex_cell(cells, self.FW_DIRECTION) == self.OUTBOUND,
                begin_time=wintimestamp(self._fw_hex_cell(cells, self.FW_BEGIN_TIME)),
                end_time=wintimestamp(self._fw_hex_cell(cells, self.FW_END_TIME)),
                repetition=self._fw_hex_cell(cells, self.FW_REPETITION),
                blocked=self._fw_hex_cell(cells, self.FW_ACTION) == self.BLOCKED,
                severity=self.SEVERITY[self._fw_hex_cell(cells, self.FW_SEVERITY)],
                rule_id=self._fw_hex_cell(cells, self.FW_RULE_ID),
                remote_host=self._fw_cell(cells, self.FW_REMOTE_HOST_NAME),
                rule_name=self._fw_cell(cells, self.FW_RULE_NAME),
                application=Path(self._fw_cell(cells, self.FW_APPLICATION)),
                user=self._fw_cell(cells, self.FW_USER),
                line_no=line_no,
                _target=self.target,
            )
        except Exception as e:
            self.target.log.warning("Error: %s on firewall log line: %d", e, line_no)

    def _line(self, line: str, line_no: int) -> Iterator[SEPLogRecord]:
        try:
            rows = csv.reader(io.StringIO(line))

            for cells in rows:
                if int(cells[self.AV_EVENT]) != self.MARKER_INFECTION:
                    return

                yield SEPLogRecord(
                    ts=self._septime(cells[self.AV_TIMESTAMP]),
                    virus=cells[self.AV_VIRUS],
                    virus_type=self.VIRUS_TYPE.get(int(cells[self.AV_VIRUS_TYPE]), "Unknown"),
                    user=cells[self.AV_USER],
                    source_file=cells[self.AV_FILE],
                    action_taken=self.ACTION.get(int(cells[self.AV_ACTION_TAKEN]), "Unknown"),
                    scan_id=cells[self.AV_SCAN_ID],
                    quarantine_id=cells[self.AV_QUARANTINE_ID],
                    virus_id=cells[self.AV_VIRUS_ID],
                    quarantined=int(cells[self.AV_QUARANTINE_STATUS]) == self.QUARANTINE_SUCCESS,
                    compressed=int(cells[self.AV_COMPRESSED]) == self.COMPRESSED,
                    depth=cells[self.AV_DEPTH],
                    still_infected=(int(cells[self.AV_STILL_INFECTED]) == self.STILL_INFECTED),
                    cleanable=int(cells[self.AV_CLEANABLE]) == self.DELETABLE,
                    deletable=int(cells[self.AV_DELETABLE]) == self.CLEANABLE,
                    confidence=cells[self.AV_CONFIDENCE],
                    prevalence=cells[self.AV_PREVALENCE],
                    risk=cells[self.AV_RISK],
                    download_url=cells[self.AV_DOWNLOADED_FROM],
                    line_no=line_no,
                    _target=self.target,
                )
        except Exception as e:
            self.target.log.warning("Error: %s on log line: %d", e, line_no)

    @export(record=SEPLogRecord)
    def logs(self) -> Iterator[SEPLogRecord]:
        """Return log records.

        Yields SEPLogRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp associated with the event.
            virus (string): Name of the virus.
            user (string): Name of the user associated with the event.
            source_file (path): File that contains the virus.
            action_taken (string): Action taken by SEP.
            virus_type (string): Description of the type of virus.
            scan_id (string): ID of the scan associated with the event.
            event_data (string): String or bytes from a virus event.
            quarantine_id (varint): ID associated with the quarantined virus.
            still_infected (boolean): Whether the system is still infected.
            quarantined (boolean): True if the virus has been quarantined succesfully.
            compressed (boolean): True if the virus was in a compressed file.
            depth (varint): How many layers of compression the virus was hidden in.
            cleanable (boolean): Whether the virus is cleanable.
            deletable (boolean): Whether the virus can be deleted.
            confidence (varint): Confidence level about threat verdict (higher is more confident).
            prevalence (varint): Prevalence of the threat (higher is more prevalent).
            risk (varint): Risk level of the threat (1-4, higher is more dangerous, 0 = unknown).
            download_url (uri): Source of the virus (if available).
            line_no (varint): Reference line number in log file.
        """
        for log_path in self.target.fs.glob(self.LOG_SEP_AV):
            with self.target.fs.path(log_path).open("rt", 0, self.codepage) as csv_file:
                line_no = 0
                while line := csv_file.readline():
                    line_no += 1
                    yield from self._line(line, line_no)

    @export(record=SEPFirewallRecord)
    def firewall(self) -> Iterator[SEPFirewallRecord]:
        """Return log firewall records.

        Yields SEPFirewallRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp associated with the event.
            protocol (string): Protocol name associated with the firewall record.
            local_ip ("net.ipaddress"): Local IP address associated with the event.
            remote_ip ("net.ipaddress"): Remote IP address associated with the event.
            local_ip6 ("net.ipaddress"): Local IPv6 address associated with the event.
            remote_ip6 ("net.ipaddress"): Remote IPv6 address associated with the event.
            local_port (varint): Local port associated with the event.
            remote_port (varint): Local port associated with the event.
            outbound (boolean): True in case of outbound traffic/connection.
            begin_time (datetime): Start of the event.
            end_time (datetime): End of the event.
            repetition (varint): How many times this event happened within the time frame.
            blocked (boolean): Whether the traffic/connection was succesfully blocked.
            severity (string): Severity of the event.
            rule_id (varint): Firewall rule ID associated with this event.
            rule_name (string): Name of the Firewall rule associated with this event.
            remote_host (string): Name of the remote host if it can be traced.
            application (path): Application responsible for/affected by event.
            user (string): User associated with the event.
            line_no (varint): Reference line number in log file.
        """

        for log_path in self.target.fs.glob(self.LOG_SEP_NET):
            log = self.target.fs.path(log_path).open("rb")
            line_no = 0
            while line := log.readline():
                line_no += 1
                if line_no == 1:
                    continue
                yield from self._fw_line(line, line_no)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.loader import open as loader_open
from dissect.target.loaders.velociraptor import VelociraptorLoader
from dissect.target.plugins.apps.edr.velociraptor import VelociraptorPlugin
from tests._utils import absolute_path
from tests.loaders.test_velociraptor import mock_velociraptor_dir

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target

mock_velociraptor_dir  # noqa: B018


def test_windows_velociraptor(mock_velociraptor_dir: Path, target_win: Target) -> None:
    """Test that a Windows Velociraptor artefact result is correctly parsed."""
    root = mock_velociraptor_dir

    with absolute_path("_data/plugins/apps/edr/velociraptor/windows-uploads.json").open("rb") as fh:
        root.joinpath("uploads.json").write_bytes(fh.read())

    with absolute_path("_data/plugins/apps/edr/velociraptor/Windows.Memory.ProcessInfo.json").open("rb") as fh:
        root.joinpath("results/Windows.Memory.ProcessInfo.json").write_bytes(fh.read())

    loader = loader_open(root)
    assert isinstance(loader, VelociraptorLoader)

    loader.map(target_win)
    target_win.apply()

    target_win.add_plugin(VelociraptorPlugin)

    results = list(target_win.velociraptor())

    record = results[0]

    assert record.name == "Microsoft.SharePoint.exe"
    assert record.pebbaseaddress == "0x295000"
    assert record.pid == 8120
    assert (
        record.imagepathname
        == "C:\\Users\\IEUser\\AppData\\Local\\Microsoft\\OneDrive\\24.070.0407.0003\\Microsoft.SharePoint.exe"
    )
    assert record.commandline == "/silentConfig"
    assert record.currentdirectory == "C:\\Windows\\system32\\"
    assert record._desc.name == "velociraptor/windows_memory_processinfo"
    assert record.env.allusersprofile == "C:\\ProgramData"



from __future__ import annotations

import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystem import Filesystem, VirtualFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.loaderutil import (
    add_virtual_ntfs_filesystem,
    parse_path_uri,
)

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        # None objects fall through (BC)
        (None, (None, None)),
        # Path objects fall through (BC/DRY)
        (Path("path"), (Path("path"), None)),
        (TargetPath(VirtualFilesystem()), (TargetPath(VirtualFilesystem()), None)),
        # Strings get upgraded to Paths
        ("/path/to/file", (Path("/path/to/file"), None)),
        # URIs get converted to Path extracted from path part and a ParseResult
        (
            "tar:///folder/file.tar.gz",
            (Path("/folder/file.tar.gz"), urllib.parse.urlparse("tar:///folder/file.tar.gz")),
        ),
        (
            "tar://relative/folder/file.tar.gz",
            (Path("relative/folder/file.tar.gz"), urllib.parse.urlparse("tar://relative/folder/file.tar.gz")),
        ),
        ("tar://~/file.tar.gz", (Path("~/file.tar.gz").expanduser(), urllib.parse.urlparse("tar://~/file.tar.gz"))),
        # But not if the URI has a faux scheme
        ("C:\\path\\to\\file", (Path("C:\\path\\to\\file"), None)),
    ],
)
def test_parse_path_uri(path: Path | str, expected: tuple[Path | None, urllib.parse.ParseResult[str] | None]) -> None:
    assert parse_path_uri(path) == expected


@pytest.mark.parametrize(
    ("boot", "mft", "expected_logs"),
    [
        (None, None, []),
        (
            None,
            False,
            [
                "Opened NTFS filesystem from <Filesystem type=virtual> but could not find $MFT, skipping",
            ],
        ),
        (None, True, []),
        (
            False,
            None,
            [
                "Failed to load NTFS filesystem from <Filesystem type=virtual>, retrying without $Boot file",
                "Opened NTFS filesystem from <Filesystem type=virtual> but could not find $MFT, skipping",
            ],
        ),
        (
            False,
            False,
            [
                "Failed to load NTFS filesystem from <Filesystem type=virtual>, retrying without $Boot file",
                "Failed to load NTFS filesystem from <Filesystem type=virtual> without $Boot file, skipping",
            ],
        ),
        (
            False,
            True,
            [
                "Failed to load NTFS filesystem from <Filesystem type=virtual>, retrying without $Boot file",
            ],
        ),
        (
            True,
            None,
            [
                "Opened NTFS filesystem from <Filesystem type=virtual> but could not find $MFT, skipping",
            ],
        ),
        (
            True,
            False,
            [
                "Failed to load NTFS filesystem from <Filesystem type=virtual>, retrying without $Boot file",
                "Failed to load NTFS filesystem from <Filesystem type=virtual> without $Boot file, skipping",
            ],
        ),
        (True, True, []),
    ],
)
def test_virtual_ntfs_resiliency(
    boot: bool | None,
    mft: bool | None,
    expected_logs: list[str],
    target_default: Target,
    caplog: pytest.LogCaptureFixture,
) -> None:
    sentinels = {
        None: None,
        False: Mock(),
        True: Mock(),
    }

    def _try_open(fs: Filesystem, path: str) -> Mock:
        state = None
        if path == "$Boot":
            state = boot
        elif path == "$MFT":
            state = mft
        return sentinels[state]

    def NtfsFilesystem(boot: Mock | None = None, mft: Mock | None = None, **kwargs) -> Mock:
        if boot is sentinels[False] or mft is sentinels[False]:
            raise ValueError("Oopsiewoopsie")

        fake_ntfs = Mock()
        fake_ntfs.ntfs.mft = None

        if mft is sentinels[True]:
            fake_ntfs.ntfs.mft = Mock()

        return fake_ntfs

    vfs = VirtualFilesystem()
    with (
        patch("dissect.target.helpers.loaderutil._try_open", new=_try_open),
        patch("dissect.target.helpers.loaderutil.NtfsFilesystem", new=NtfsFilesystem),
    ):
        add_virtual_ntfs_filesystem(target_default, vfs)

    assert caplog.messages == expected_logs
    assert hasattr(vfs, "ntfs") == (bool(mft))



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.hypervisor.disk import hdd
from dissect.hypervisor.disk.c_hdd import c_hdd

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class HdsContainer(Container):
    __type__ = "hds"

    def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
        f = fh
        if not hasattr(fh, "read"):
            f = fh.open("rb")

        self.hds = hdd.HDS(f)
        super().__init__(fh, self.hds.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return fh.read(16) in (c_hdd.SIGNATURE_STRUCTURED_DISK_V1, c_hdd.SIGNATURE_STRUCTURED_DISK_V2)

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return path.suffix.lower() == ".hds"

    def read(self, length: int) -> bytes:
        return self.hds.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.hds.seek(offset, whence)

    def tell(self) -> int:
        return self.hds.tell()

    def close(self) -> None:
        self.hds.close()



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.unix.bsd.freebsd._os import FreeBsdPlugin
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.plugin import OSPlugin
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("data_path", "target_path", "expected_version", "os_plugin"),
    [
        (
            "_data/plugins/os/unix/linux/debian/_os/debian-os-release",
            "/etc/os-release",
            "Debian GNU/Linux 11 (bullseye)",
            LinuxPlugin,
        ),
        (
            "_data/plugins/os/unix/linux/debian/_os/ubuntu-os-release",
            "/etc/os-release",
            "Ubuntu 22.04.2 LTS (Jammy Jellyfish)",
            LinuxPlugin,
        ),
        (
            "_data/plugins/os/unix/linux/redhat/_os/centos-os-release",
            "/etc/os-release",
            "CentOS Linux 8",
            LinuxPlugin,
        ),
        (
            "_data/plugins/os/unix/linux/redhat/_os/fedora-os-release",
            "/etc/os-release",
            "Fedora Linux 37 (Container Image)",
            LinuxPlugin,
        ),
        (
            "_data/plugins/os/unix/linux/suse/_os/opensuse-os-release",
            "/etc/os-release",
            "openSUSE Leap 15.4",
            LinuxPlugin,
        ),
        (
            "_data/plugins/os/unix/linux/debian/_os/ubuntu-lsb-release",
            "/etc/lsb-release",
            "Ubuntu 22.04.2 LTS",
            LinuxPlugin,
        ),
        (
            "_data/plugins/os/unix/bsd/freebsd/_os/freebsd-freebsd-version",
            "/bin/freebsd-version",
            "13.0-RELEASE",
            FreeBsdPlugin,
        ),
        (
            "_data/plugins/os/unix/linux/alpine/_os/alpine-os-release",
            "/etc/os-release",
            "Alpine Linux 3.18.4",
            LinuxPlugin,
        ),
    ],
)
def test_unix_version_detection(
    target_unix: Target,
    fs_unix: VirtualFilesystem,
    data_path: str,
    target_path: str,
    expected_version: str,
    os_plugin: type[OSPlugin],
) -> None:
    fs_unix.map_file(target_path, absolute_path(data_path))
    target_unix.add_plugin(os_plugin)

    assert target_unix.version == expected_version


@pytest.mark.parametrize(
    ("content", "target_path", "os_plugin"),
    [
        ("Fedora release 37 (Thirty Seven)", "/etc/fedora-release", LinuxPlugin),
        ("CentOS Linux release 8.4.2105", "/etc/centos-release", LinuxPlugin),
    ],
)
def test_unix_version_detection_short(
    target_unix: Target, fs_unix: VirtualFilesystem, content: str, target_path: str, os_plugin: type[OSPlugin]
) -> None:
    fs_unix.map_file_fh(target_path, BytesIO(content.encode()))
    target_unix.add_plugin(os_plugin)

    assert target_unix.version == content


def test_unix_os_release_directory_regression(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh(
        "/etc/upstream-release/lsb-release",
        BytesIO(
            b'DISTRIB_ID=Ubuntu\nDISTRIB_RELEASE=16.04\nDISTRIB_CODENAME=xenial\nDISTRIB_DESCRIPTION="Ubuntu 16.04 LTS"'
        ),
    )
    fs_unix.map_file_fh(
        "/etc/lsb-release",
        BytesIO(
            b'DISTRIB_ID=LinuxMint\nDISTRIB_RELEASE=19\nDISTRIB_CODENAME=tara\nDISTRIB_DESCRIPTION="Linux Mint 19 Tara"'
        ),
    )
    target_unix.add_plugin(LinuxPlugin)
    assert target_unix.version == "Linux Mint 19 Tara"



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from dissect.target.helpers import keychain
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import NamespacePlugin

if TYPE_CHECKING:
    from dissect.target.target import Target

GENERIC_DOWNLOAD_RECORD_FIELDS = [
    ("datetime", "ts_start"),
    ("datetime", "ts_end"),
    ("string", "browser"),
    ("varint", "id"),
    ("path", "path"),
    ("uri", "url"),
    ("filesize", "size"),
    ("string", "state"),
    ("path", "source"),
]

GENERIC_EXTENSION_RECORD_FIELDS = [
    ("datetime", "ts_install"),
    ("datetime", "ts_update"),
    ("string", "browser"),
    ("string", "extension_id"),
    ("string", "name"),
    ("string", "short_name"),
    ("string", "default_title"),
    ("string", "description"),
    ("string", "version"),
    ("path", "ext_path"),
    ("boolean", "from_webstore"),
    ("string[]", "permissions"),
    ("varint", "manifest_version"),
    ("path", "source"),
]

GENERIC_COOKIE_FIELDS = [
    ("datetime", "ts_created"),
    ("datetime", "ts_last_accessed"),
    ("string", "browser"),
    ("string", "name"),
    ("string", "value"),
    ("string", "host"),
    ("string", "path"),
    ("varint", "expiry"),
    ("boolean", "is_secure"),
    ("boolean", "is_http_only"),
    ("boolean", "same_site"),
    ("path", "source"),
]

GENERIC_HISTORY_RECORD_FIELDS = [
    ("datetime", "ts"),
    ("string", "browser"),
    ("varint", "id"),
    ("uri", "url"),
    ("string", "title"),
    ("string", "description"),
    ("string", "host"),
    ("varint", "visit_type"),
    ("varint", "visit_count"),
    ("boolean", "hidden"),
    ("boolean", "typed"),
    ("varint", "session"),
    ("varint", "from_visit"),
    ("uri", "from_url"),
    ("path", "source"),
]

GENERIC_PASSWORD_RECORD_FIELDS = [
    ("datetime", "ts_created"),
    ("datetime", "ts_last_used"),
    ("datetime", "ts_last_changed"),
    ("string", "browser"),
    ("varint", "id"),
    ("uri", "url"),
    ("bytes", "encrypted_username"),
    ("bytes", "encrypted_password"),
    ("bytes", "encrypted_notes"),
    ("string", "decrypted_username"),
    ("string", "decrypted_password"),
    ("string", "decrypted_notes"),
    ("path", "source"),
]

BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "browser/download", GENERIC_DOWNLOAD_RECORD_FIELDS
)
BrowserExtensionRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "browser/extension", GENERIC_EXTENSION_RECORD_FIELDS
)
BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "browser/history", GENERIC_HISTORY_RECORD_FIELDS
)
BrowserCookieRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "browser/cookie", GENERIC_COOKIE_FIELDS
)
BrowserPasswordRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "browser/password", GENERIC_PASSWORD_RECORD_FIELDS
)


class BrowserPlugin(NamespacePlugin):
    __namespace__ = "browser"

    def __init__(self, target: Target):
        super().__init__(target)
        self.keychain = cache(self.keychain)

    def keychain(self) -> set:
        """Retrieve a set of passphrases to use for decrypting saved browser credentials.

        Always adds an empty passphrase as some browsers encrypt values using empty passphrases.

        Returns:
            Set of passphrase strings.
        """
        passphrases = set()
        for provider in [self.__namespace__, "browser", "user", None]:
            for key in keychain.get_keys_for_provider(provider) if provider else keychain.get_keys_without_provider():
                if key.key_type == keychain.KeyType.PASSPHRASE:
                    passphrases.add(key.value)

        passphrases.add("")
        return passphrases


def try_idna(url: str) -> bytes:
    """Attempts to convert a possible Unicode url to ASCII using the IDNA standard.

    Args:
        url: A String containing the url to be converted.

    Returns: Bytes object with the ASCII version of the url.
    """
    try:
        return url.encode("idna")
    except Exception:
        return url



from __future__ import annotations

import stat
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.filesystems.nfs import NfsFilesystem, NfsFilesystemEntry, NfsStream
from dissect.target.helpers.nfs.nfs3 import (
    EntryPlus,
    FileAttributes,
    FileHandle,
    FileType,
    NfsTime,
    SpecData,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def mock_nfs_client() -> Iterator[MagicMock]:
    with patch("dissect.target.helpers.nfs.client.nfs.Client", autospec=True) as mock_client:
        yield mock_client


@pytest.fixture
def nfs_filesystem(mock_nfs_client: MagicMock) -> NfsFilesystem:
    client_factory = MagicMock(return_value=mock_nfs_client)
    root_handle = FileHandle(opaque=b"root_handle")
    return NfsFilesystem(client_factory, root_handle)


@pytest.fixture
def nfs_filesystem_entry(nfs_filesystem: NfsFilesystem) -> NfsFilesystemEntry:
    file_handle = FileHandle(opaque=b"file_handle")
    attributes = FileAttributes(
        type=FileType.REG,
        mode=0o644,
        nlink=1,
        uid=1000,
        gid=1000,
        size=1024,
        used=1024,
        rdev=SpecData(8, 9),
        fsid=0,
        fileid=0,
        atime=NfsTime(1, 2),
        mtime=NfsTime(3, 4),
        ctime=NfsTime(5, 6),
    )
    return NfsFilesystemEntry(nfs_filesystem, "/file", file_handle, attributes)


def test_get_root(nfs_filesystem: NfsFilesystem) -> None:
    entry = nfs_filesystem.get("/")
    assert isinstance(entry, NfsFilesystemEntry)
    assert entry.path == "/"
    assert entry.entry.opaque == b"root_handle"


def test_get_subdirectory(nfs_filesystem: NfsFilesystem, mock_nfs_client: MagicMock) -> None:
    mock_nfs_client.lookup.return_value = MagicMock(
        object=FileHandle(opaque=b"subdir_handle"), obj_attributes=MagicMock()
    )
    entry = nfs_filesystem.get("/subdir")
    mock_nfs_client.lookup.assert_called_with("subdir", FileHandle(opaque=b"root_handle"))
    assert isinstance(entry, NfsFilesystemEntry)
    assert entry.path == "subdir"
    assert entry.entry.opaque == b"subdir_handle"


def test_get_from_entry(mock_nfs_client: MagicMock, nfs_filesystem_entry: NfsFilesystemEntry) -> None:
    nfs_filesystem_entry._attributes.type = FileType.DIR
    mock_nfs_client.lookup.return_value = MagicMock(
        object=FileHandle(opaque=b"subdir_handle"), obj_attributes=MagicMock()
    )
    entry = nfs_filesystem_entry.get("subdir")
    mock_nfs_client.lookup.assert_called_with("subdir", nfs_filesystem_entry.entry)
    assert isinstance(entry, NfsFilesystemEntry)
    assert entry.path == "subdir"
    assert entry.entry.opaque == b"subdir_handle"


def test_is_file(nfs_filesystem_entry: NfsFilesystemEntry) -> None:
    nfs_filesystem_entry._attributes.type = FileType.REG
    assert nfs_filesystem_entry.is_file()


def test_is_dir(nfs_filesystem_entry: NfsFilesystemEntry) -> None:
    nfs_filesystem_entry._attributes.type = FileType.DIR
    assert nfs_filesystem_entry.is_dir()


def test_is_symlink(nfs_filesystem_entry: NfsFilesystemEntry) -> None:
    nfs_filesystem_entry._attributes.type = FileType.LNK
    assert nfs_filesystem_entry.is_symlink()


def test_readlink(nfs_filesystem_entry: NfsFilesystemEntry, mock_nfs_client: MagicMock) -> None:
    mock_nfs_client.readlink.return_value = "/target"
    target = nfs_filesystem_entry.readlink()
    mock_nfs_client.readlink.assert_called_with(FileHandle(opaque=b"file_handle"))
    assert target == "/target"


def test_iterdir(nfs_filesystem_entry: NfsFilesystemEntry, mock_nfs_client: MagicMock) -> None:
    nfs_filesystem_entry._attributes.type = FileType.DIR
    mock_nfs_client.readdir.return_value = MagicMock(
        entries=[
            EntryPlus(
                fileid=1,
                cookie=1,
                name="file1",
                handle=FileHandle(opaque=b"file1_handle"),
                attributes=nfs_filesystem_entry._attributes,
            ),
            EntryPlus(
                fileid=2,
                cookie=2,
                name="file2",
                handle=FileHandle(opaque=b"file2_handle"),
                attributes=nfs_filesystem_entry._attributes,
            ),
        ]
    )
    entries = list(nfs_filesystem_entry.iterdir())
    mock_nfs_client.readdir.assert_called_with(FileHandle(opaque=b"file_handle"))
    assert entries == ["file1", "file2"]


def test_stat(nfs_filesystem_entry: NfsFilesystemEntry) -> None:
    stat_result = nfs_filesystem_entry.stat()
    assert stat_result.st_mode == nfs_filesystem_entry._attributes.mode | stat.S_IFREG
    assert stat_result.st_uid == nfs_filesystem_entry._attributes.uid
    assert stat_result.st_gid == nfs_filesystem_entry._attributes.gid
    assert stat_result.st_size == nfs_filesystem_entry._attributes.size
    assert stat_result.st_atime == nfs_filesystem_entry._attributes.atime.seconds
    assert stat_result.st_mtime == nfs_filesystem_entry._attributes.mtime.seconds
    assert stat_result.st_ctime == nfs_filesystem_entry._attributes.ctime.seconds
    assert stat_result.st_atime_ns == nfs_filesystem_entry._attributes.atime.nseconds
    assert stat_result.st_mtime_ns == nfs_filesystem_entry._attributes.mtime.nseconds
    assert stat_result.st_ctime_ns == nfs_filesystem_entry._attributes.ctime.nseconds


def test_open(nfs_filesystem_entry: NfsFilesystemEntry, mock_nfs_client: MagicMock) -> None:
    stream = nfs_filesystem_entry.open()
    assert isinstance(stream, NfsStream)
    assert stream._client == mock_nfs_client
    assert stream._file_handle.opaque == nfs_filesystem_entry.entry.opaque
    assert stream.size == nfs_filesystem_entry._attributes.size


def test_stream_read(nfs_filesystem_entry: NfsFilesystemEntry, mock_nfs_client: MagicMock) -> None:
    mock_nfs_client.readfile.return_value = iter([b"hello", b" ", b"world"])
    stream = nfs_filesystem_entry.open()
    data = stream.read()
    assert data == b"hello world"
    mock_nfs_client.readfile.assert_called_with(FileHandle(opaque=b"file_handle"), 0, 1024 * 1024)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.keyprovider import (
    KeyProviderPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class CredHistKeyProviderPlugin(KeyProviderPlugin):
    """Windows CREDHIST SHA1-hash key provider plugin."""

    __namespace__ = "dpapi.keyprovider.credhist"

    def check_compatible(self) -> None:
        if not self.target.has_function("credhist"):
            raise UnsupportedPluginError("CREDHIST plugin not available on target")

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield Windows CREDHIST SHA1 hashes."""
        for credhist in self.target.credhist():
            if value := credhist.sha1:
                yield self.__namespace__, value



from __future__ import annotations

from pathlib import Path

from dissect.target.filesystems.cpio import CpioFilesystem
from tests._utils import absolute_path


def test_cpio_uncompressed() -> None:
    cpio_path = Path(absolute_path("_data/filesystems/cpio/initrd.img-6.1.0-17-amd64"))

    with cpio_path.open("rb") as fh:
        assert CpioFilesystem.detect(fh)

        fs = CpioFilesystem(fh)
        assert [f.name for f in fs.path("/").iterdir()] == ["kernel"]


def test_cpio_compressed_zstd() -> None:
    cpio_path = Path(absolute_path("_data/filesystems/cpio/initrd.img-6.1.0-15-amd64"))

    with cpio_path.open("rb") as fh:
        assert CpioFilesystem.detect(fh)

        fs = CpioFilesystem(fh)
        assert [f.name for f in fs.path("/").iterdir()] == [
            "bin",
            "conf",
            "etc",
            "init",
            "lib",
            "lib64",
            "run",
            "sbin",
            "scripts",
            "usr",
        ]



from __future__ import annotations

import re
from enum import IntEnum
from functools import lru_cache
from typing import TYPE_CHECKING

from dissect.util.ts import wintimestamp

from dissect.target.exceptions import (
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
)
from dissect.target.helpers.record import WindowsInterfaceRecord
from dissect.target.plugins.os.default.network import NetworkPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.target import Target


class IfTypes(IntEnum):
    OTHER = 1
    REGULAR_1822 = 2
    HDH_1822 = 3
    DDN_X25 = 4
    RFC877_X25 = 5
    ETHERNET_CSMACD = 6
    IS088023_CSMACD = 7
    ISO88024_TOKENBUS = 8
    ISO88025_TOKENRING = 9
    ISO88026_MAN = 10
    STARLAN = 11
    PROTEON_10MBIT = 12
    PROTEON_80MBIT = 13
    HYPERCHANNEL = 14
    FDDI = 15
    LAP_B = 16
    SDLC = 17
    DS1 = 18
    E1 = 19
    BASIC_ISDN = 20
    PRIMARY_ISDN = 21
    PROP_POINT2POINT_SERIAL = 22
    PPP = 23
    SOFTWARE_LOOPBACK = 24
    EON = 25
    ETHERNET_3MBIT = 26
    NSIP = 27
    SLIP = 28
    ULTRA = 29
    DS3 = 30
    SIP = 31
    FRAMERELAY = 32
    RS232 = 33
    PARA = 34
    ARCNET = 35
    ARCNET_PLUS = 36
    ATM = 37
    MIO_X25 = 38
    SONET = 39
    X25_PLE = 40
    ISO88022_LLC = 41
    LOCALTALK = 42
    SMDS_DXI = 43
    FRAMERELAY_SERVICE = 44
    V35 = 45
    HSSI = 46
    HIPPI = 47
    MODEM = 48
    AAL5 = 49
    SONET_PATH = 50
    SONET_VT = 51
    SMDS_ICIP = 52
    PROP_VIRTUAL = 53
    PROP_MULTIPLEXOR = 54
    IEEE80212 = 55
    FIBRECHANNEL = 56
    HIPPIINTERFACE = 57
    FRAMERELAY_INTERCONNECT = 58
    AFLANE_8023 = 59
    AFLANE_8025 = 60
    CCTEMUL = 61
    FASTETHER = 62
    ISDN = 63
    V11 = 64
    V36 = 65
    G703_64K = 66
    G703_2MB = 67
    QLLC = 68
    FASTETHER_FX = 69
    CHANNEL = 70
    IEEE80211 = 71
    IBM370PARCHAN = 72
    ESCON = 73
    DLSW = 74
    ISDN_S = 75
    ISDN_U = 76
    LAP_D = 77
    IPSWITCH = 78
    RSRB = 79
    ATM_LOGICAL = 80
    DS0 = 81
    DS0_BUNDLE = 82
    BSC = 83
    ASYNC = 84
    CNR = 85
    ISO88025R_DTR = 86
    EPLRS = 87
    ARAP = 88
    PROP_CNLS = 89
    HOSTPAD = 90
    TERMPAD = 91
    FRAMERELAY_MPI = 92
    X213 = 93
    ADSL = 94
    RADSL = 95
    SDSL = 96
    VDSL = 97
    ISO88025_CRFPRINT = 98
    MYRINET = 99
    VOICE_EM = 100
    VOICE_FXO = 101
    VOICE_FXS = 102
    VOICE_ENCAP = 103
    VOICE_OVERIP = 104
    ATM_DXI = 105
    ATM_FUNI = 106
    ATM_IMA = 107
    PPPMULTILINKBUNDLE = 108
    IPOVER_CDLC = 109
    IPOVER_CLAW = 110
    STACKTOSTACK = 111
    VIRTUALIPADDRESS = 112
    MPC = 113
    IPOVER_ATM = 114
    ISO88025_FIBER = 115
    TDLC = 116
    GIGABITETHERNET = 117
    HDLC = 118
    LAP_F = 119
    V37 = 120
    X25_MLP = 121
    X25_HUNTGROUP = 122
    TRANSPHDLC = 123
    INTERLEAVE = 124
    FAST = 125
    IP = 126
    DOCSCABLE_MACLAYER = 127
    DOCSCABLE_DOWNSTREAM = 128
    DOCSCABLE_UPSTREAM = 129
    A12MPPSWITCH = 130
    TUNNEL = 131
    COFFEE = 132
    CES = 133
    ATM_SUBINTERFACE = 134
    L2_VLAN = 135
    L3_IPVLAN = 136
    L3_IPXVLAN = 137
    DIGITALPOWERLINE = 138
    MEDIAMAILOVERIP = 139
    DTM = 140
    DCN = 141
    IPFORWARD = 142
    MSDSL = 143
    IEEE1394 = 144
    IF_GSN = 145
    DVBRCC_MACLAYER = 146
    DVBRCC_DOWNSTREAM = 147
    DVBRCC_UPSTREAM = 148
    ATM_VIRTUAL = 149
    MPLS_TUNNEL = 150
    SRP = 151
    VOICEOVERATM = 152
    VOICEOVERFRAMERELAY = 153
    IDSL = 154
    COMPOSITELINK = 155
    SS7_SIGLINK = 156
    PROP_WIRELESS_P2P = 157
    FR_FORWARD = 158
    RFC1483 = 159
    USB = 160
    IEEE8023AD_LAG = 161
    BGP_POLICY_ACCOUNTING = 162
    FRF16_MFR_BUNDLE = 163
    H323_GATEKEEPER = 164
    H323_PROXY = 165
    MPLS = 166
    MF_SIGLINK = 167
    HDSL2 = 168
    SHDSL = 169
    DS1_FDL = 170
    POS = 171
    DVB_ASI_IN = 172
    DVB_ASI_OUT = 173
    PLC = 174
    NFAS = 175
    TR008 = 176
    GR303_RDT = 177
    GR303_IDT = 178
    ISUP = 179
    PROP_DOCS_WIRELESS_MACLAYER = 180
    PROP_DOCS_WIRELESS_DOWNSTREAM = 181
    PROP_DOCS_WIRELESS_UPSTREAM = 182
    HIPERLAN2 = 183
    PROP_BWA_P2MP = 184
    SONET_OVERHEAD_CHANNEL = 185
    DIGITAL_WRAPPER_OVERHEAD_CHANNEL = 186
    AAL2 = 187
    RADIO_MAC = 188
    ATM_RADIO = 189
    IMT = 190
    MVL = 191
    REACH_DSL = 192
    FR_DLCI_ENDPT = 193
    ATM_VCI_ENDPT = 194
    OPTICAL_CHANNEL = 195
    OPTICAL_TRANSPORT = 196
    WWANPP = 243
    WWANPP2 = 244


def _try_value(subkey: RegistryKey, value: str) -> str | list | None:
    try:
        return subkey.value(value).value
    except RegistryValueNotFoundError:
        return None


def _get_config_value(key: RegistryKey, name: str, sep: str | None = None) -> set:
    value = _try_value(key, name)
    if not value or value in ("", "0.0.0.0", None, [], ["0.0.0.0"]):
        return set()
    if sep and isinstance(value, str):
        re_sep = "|".join(map(re.escape, sep))
        value = re.split(re_sep, value)
    if isinstance(value, list):
        return set(value)

    return {value}


def _construct_interface(key: RegistryKey, ip_key: str, subnet_key: str) -> set[str]:
    interface = ""

    if ip := _get_config_value(key, ip_key):
        interface = next(iter(ip))

    if not interface:
        return set()

    if subnet := _get_config_value(key, subnet_key):
        interface = f"{interface}/{next(iter(subnet))}"

    if not interface:
        return set()

    return {interface}


class WindowsNetworkPlugin(NetworkPlugin):
    """Windows network interface plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._extract_network_device_config = lru_cache(128)(self._extract_network_device_config)

    def _interfaces(self) -> Iterator[WindowsInterfaceRecord]:
        """Yields found Windows interfaces used by :meth:`NetworkPlugin.interfaces() <dissect.target.plugins.general.network.NetworkPlugin.interfaces>`."""  # noqa: E501

        # Get all the network interfaces
        for key in self.target.registry.keys(
            "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Class\\{4d36e972-e325-11ce-bfc1-08002be10318}"
        ):
            for subkey in key.subkeys():
                device_info = {}

                if (net_cfg_instance_id := _try_value(subkey, "NetCfgInstanceId")) is None:
                    # if no NetCfgInstanceId is found, skip this network interface
                    continue

                # Extract the network device configuration for given interface id
                if not (config := self._extract_network_device_config(net_cfg_instance_id)):
                    continue

                # Extract a network device name for given interface id
                try:
                    name_key = self.target.registry.key(
                        f"HKLM\\SYSTEM\\CurrentControlSet\\Control\\Network\\{{4D36E972-E325-11CE-BFC1-08002BE10318}}\\{net_cfg_instance_id}\\Connection"
                    )
                    if value_name := _try_value(name_key, "Name"):
                        device_info["name"] = value_name
                except RegistryKeyNotFoundError:
                    pass

                # Extract the metric value from the interface registry key
                try:
                    interface_key = self.target.registry.key(
                        f"HKLM\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\{net_cfg_instance_id}"
                    )
                    if value_metric := _try_value(interface_key, "InterfaceMetric"):
                        device_info["metric"] = value_metric
                except RegistryKeyNotFoundError:
                    pass

                # Extract the rest of the device information
                if mac_address := _try_value(subkey, "NetworkAddress"):
                    device_info["mac"] = [mac_address]
                device_info["vlan"] = _try_value(subkey, "VlanID")

                if timestamp := _try_value(subkey, "NetworkInterfaceInstallTimestamp"):
                    device_info["first_connected"] = wintimestamp(timestamp)

                if type_device := _try_value(subkey, "*IfType"):
                    device_info["type"] = IfTypes(int(type_device)).name

                # Yield a record for each non-empty configuration
                for conf in config:
                    # If no configuration is found or all configurations are empty,
                    # skip this network interface.
                    if not conf or not any(
                        [
                            conf["cidr"],
                            conf["gateway"],
                            conf["dns"],
                            conf["search_domain"],
                        ]
                    ):
                        continue

                    # Create a copy of device_info to avoid overwriting
                    record_info = device_info.copy()
                    record_info.update(conf)
                    yield WindowsInterfaceRecord(
                        **record_info,
                        source=f"HKLM\\SYSTEM\\{subkey.path}",
                        _target=self.target,
                    )

    def _extract_network_device_config(self, interface_id: str) -> list[dict[str, set | bool | None]]:
        """Extract network device configuration from the given interface_id for all ControlSets on the system."""

        dhcp_config = {
            "cidr": set(),
            "gateway": set(),
            "dns": set(),
            "search_domain": set(),
        }

        static_config = {
            "cidr": set(),
            "gateway": set(),
            "dns": set(),
            "search_domain": set(),
        }

        # Get the registry keys for the given interface id
        try:
            keys = list(
                self.target.registry.keys(
                    f"HKLM\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\{interface_id}"
                )
            )
        except RegistryKeyNotFoundError:
            return []

        if not keys:
            return []

        for key in keys:
            # Extract DHCP configuration from the registry
            dhcp_config["cidr"].update(_construct_interface(key, "DhcpIPAddress", "DhcpSubnetMask"))
            dhcp_config["gateway"].update(_get_config_value(key, "DhcpDefaultGateway"))
            dhcp_config["dns"].update(_get_config_value(key, "DhcpNameServer", " ,"))
            dhcp_config["search_domain"].update(_get_config_value(key, "DhcpDomain"))

            # Extract static configuration from the registry
            static_config["cidr"].update(_construct_interface(key, "IPAddress", "SubnetMask"))
            static_config["gateway"].update(_get_config_value(key, "DefaultGateway"))
            static_config["dns"].update(_get_config_value(key, "NameServer", " ,"))
            static_config["search_domain"].update(_get_config_value(key, "Domain"))

        dhcp_config["enabled"] = _try_value(key, "EnableDHCP") == 1
        dhcp_config["dhcp"] = True

        static_config["enabled"] = None
        static_config["dhcp"] = False

        # Return both configurations
        return [dhcp_config, static_config]



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.cache import Cache
from dissect.target.plugins.os.windows.amcache import AmcachePlugin
from dissect.target.plugins.os.windows.ual import UalPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_cache_namespace(target_bare: Target) -> None:
    cache1 = Cache(AmcachePlugin.__call__, cls=AmcachePlugin)
    cache2 = Cache(UalPlugin.__call__, cls=UalPlugin)

    assert cache1.fname != cache2.fname
    assert cache1.fname == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.__call__"
    assert cache2.fname == "dissect.target.plugins.os.windows.ual.UalPlugin.__call__"

    target_bare._config.CACHE_DIR = "/tmp"
    assert (
        cache1.cache_path(target_bare, ("a", 1234)).name
        == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.__call__.KCdhJywgMTIzNCk=.zstd"
    )
    assert (
        cache2.cache_path(target_bare, ("b", 5678)).name
        == "dissect.target.plugins.os.windows.ual.UalPlugin.__call__.KCdiJywgNTY3OCk=.zstd"
    )


def test_cache_filename(target_win: Target) -> None:
    plugin1 = AmcachePlugin(target_win)
    plugin2 = UalPlugin(target_win)

    cache1 = Cache(AmcachePlugin.applications)
    cache2 = Cache(UalPlugin.client_access)
    cache3 = Cache(plugin1.applications)
    cache4 = Cache(plugin2.client_access)

    assert cache1.fname == cache3.fname == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.applications"
    assert cache2.fname == cache4.fname == "dissect.target.plugins.os.windows.ual.UalPlugin.client_access"

    target_win._config.CACHE_DIR = "/tmp"
    assert (
        cache1.cache_path(target_win, ()).name
        == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.applications.KCk=.zstd"
    )
    assert (
        cache2.cache_path(target_win, ()).name
        == "dissect.target.plugins.os.windows.ual.UalPlugin.client_access.KCk=.zstd"
    )
    assert (
        cache3.cache_path(target_win, ()).name
        == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.applications.KCk=.zstd"
    )
    assert (
        cache4.cache_path(target_win, ()).name
        == "dissect.target.plugins.os.windows.ual.UalPlugin.client_access.KCk=.zstd"
    )
